import dataclasses
import hashlib
import os
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import timm
import torch
from torch.hub import load_state_dict_from_url
import yaml

from .inceptionv4 import inceptionv4 as _inceptionv4
from .inceptionv4_no_batchnorm import inceptionv4 as _inceptionv4_no_bn
from .resnet_preact import resnet34_preact as _resnet34_preact
from .vgg16mod import vgg16mod as _vgg16mod
from .transforms import PatchClassification


class WsinferException(Exception):
    "Base class for wsinfer exceptions."


class UnknownArchitectureError(WsinferException):
    """Architecture is unknown and cannot be found."""


class ModelWeightsNotFound(WsinferException):
    """Model weights are not found, likely because they are not in the registry."""


class DuplicateModelWeights(WsinferException):
    """A duplicate key was passed to the model weights registry."""


class ModelRegistrationError(WsinferException):
    """Error during model registration."""


PathType = Union[str, Path]


def _sha256sum(path: PathType) -> str:
    """Calculate SHA256 of a file."""
    sha = hashlib.sha256()
    with open(path, "rb") as f:
        while True:
            data = f.read(1024 * 64)  # 64 kb
            if not data:
                break
            sha.update(data)
    return sha.hexdigest()


@dataclasses.dataclass
class Weights:
    """Container for data associated with a trained model."""

    name: str
    architecture: str
    num_classes: int
    transform: PatchClassification
    patch_size_pixels: int
    spacing_um_px: float
    class_names: List[str]
    url: Optional[str] = None
    url_file_name: Optional[str] = None
    file: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        if len(set(self.class_names)) != len(self.class_names):
            raise ValueError("class_names cannot contain duplicates")
        if len(self.class_names) != self.num_classes:
            raise ValueError("length of class_names must be equal to num_classes")

    @staticmethod
    def _validate_input(d) -> None:
        """Raise error if invalid input."""

        if not isinstance(d, dict):
            raise ValueError("expected config to be a dictionary")

        # Validate contents.
        # Validate keys.
        required_keys = [
            "version",
            "name",
            "architecture",
            "num_classes",
            "transform",
            "patch_size_pixels",
            "spacing_um_px",
            "class_names",
        ]
        optional_keys = ["url", "url_file_name", "file", "metadata"]
        all_keys = required_keys + optional_keys
        for req_key in required_keys:
            if req_key not in d.keys():
                raise KeyError(f"required key not found: '{req_key}'")
        unknown_keys = [k for k in d.keys() if k not in all_keys]
        if unknown_keys:
            raise KeyError(f"unknown keys: {unknown_keys}")
        for req_key in ["resize_size", "mean", "std"]:
            if req_key not in d["transform"].keys():
                raise KeyError(
                    f"required key not found in 'transform' section: '{req_key}'"
                )

        # We include a 'version' key so we can handle updates if needed in the future.
        # At this point, we only support version 1.0.
        if d["version"] != "1.0":
            raise ValueError("config file must include version: '1.0'.")
        # Either 'url' or 'file' is required. If 'url' is used, then 'url_file_name' is
        # required.
        if "url" not in d.keys() and "file" not in d.keys():
            raise KeyError("'url' or 'file' must be provided")
        if "url" in d.keys() and "file" in d.keys():
            raise KeyError("only on of 'url' and 'file' can be used")
        if "url" in d.keys() and "url_file_name" not in d.keys():
            raise KeyError("when using 'url', 'url_file_name' must also be provided")

        # Validate types.
        if not isinstance("architecture", str):
            raise ValueError("'architecture' must be a string")
        if not isinstance("name", str):
            raise ValueError("'name' must be a string")
        if "url" in d.keys() and not isinstance(d["url"], str):
            raise ValueError("'url' must be a string")
        if "url_file_name" in d.keys() and not isinstance(d["url"], str):
            raise ValueError("'url_file_name' must be a string")
        if not isinstance(d["num_classes"], int):
            raise ValueError("'num_classes' must be an integer")
        if not isinstance(d["transform"]["resize_size"], int):
            raise ValueError("'transform.resize_size' must be an integer")
        if not isinstance(d["transform"]["mean"], list):
            raise ValueError("'transform.mean' must be a list")
        if not all(isinstance(num, float) for num in d["transform"]["mean"]):
            raise ValueError("'transform.mean' must be a list of floats")
        if not isinstance(d["transform"]["std"], list):
            raise ValueError("'transform.std' must be a list")
        if not all(isinstance(num, float) for num in d["transform"]["std"]):
            raise ValueError("'transform.std' must be a list of floats")
        if not isinstance(d["patch_size_pixels"], int) or d["patch_size_pixels"] <= 0:
            raise ValueError("patch_size_pixels must be a positive integer")
        if not isinstance(d["spacing_um_px"], float) or d["spacing_um_px"] <= 0:
            raise ValueError("spacing_um_px must be a positive float")
        if not isinstance(d["class_names"], list):
            raise ValueError("'class_names' must be a list")
        if not all(isinstance(c, str) for c in d["class_names"]):
            raise ValueError("'class_names' must be a list of strings")

        # Validate values.
        if len(d["transform"]["mean"]) != 3:
            raise ValueError("transform.mean must be a list of three numbers")
        if len(d["transform"]["std"]) != 3:
            raise ValueError("transform.std must be a list of three numbers")
        if len(d["class_names"]) != len(set(d["class_names"])):
            raise ValueError("duplicate values found in 'class_names'")
        if len(d["class_names"]) != d["num_classes"]:
            raise ValueError("mismatch between length of class_names and num_classes.")
        if "file" in d.keys():
            file = Path(path).parent / d["file"]
            file = file.resolve()
            if not file.exists():
                raise FileNotFoundError(f"'file' not found: {file}")

    @classmethod
    def from_yaml(cls, path):
        """Create a new instance of Weights from a YAML file."""

        with open(path) as f:
            d = yaml.safe_load(f)
        cls._validate_input(d)

        transform = PatchClassification(
            resize_size=d["transform"]["resize_size"],
            mean=d["transform"]["mean"],
            std=d["transform"]["std"],
        )
        return Weights(
            name=d["name"],
            architecture=d["architecture"],
            url=d.get("url"),
            url_file_name=d.get("url_file_name"),
            file=d.get("file"),
            num_classes=d["num_classes"],
            transform=transform,
            patch_size_pixels=d["patch_size_pixels"],
            spacing_um_px=d["spacing_um_px"],
            class_names=d["class_names"],
        )

    def load_model(self) -> torch.nn.Module:
        """Return the pytorch implementation of the architecture with weights loaded."""
        model = _create_model(name=self.architecture, num_classes=self.num_classes)

        # Load state dict.
        if self.url and self.url_file_name:
            state_dict = load_state_dict_from_url(
                url=self.url,
                map_location="cpu",
                check_hash=True,
                file_name=self.url_file_name,
            )
        elif self.file:
            state_dict = torch.load(self.file, map_location="cpu")
            # When training with timm scripts, weights are saved in 'state_dict' key.
            if "state_dict" in state_dict.keys():
                state_dict = state_dict["state_dict"]
        else:
            raise RuntimeError("cannot find weights")

        model.load_state_dict(state_dict, strict=True)
        return model

    def get_sha256_of_weights(self) -> str:
        """Return the sha256 of the weights file."""
        if self.url and self.url_file_name:
            p = Path(torch.hub.get_dir()) / "checkpoints" / self.url_file_name
        elif self.file:
            p = Path(self.file)
        else:
            raise RuntimeError("cannot find path to weights")
        sha = _sha256sum(p)
        return sha


# Container for all models we can use that are not in timm.
_model_registry: Dict[str, Callable[[int], torch.nn.Module]] = {
    "inceptionv4": _inceptionv4,
    "inceptionv4nobn": _inceptionv4_no_bn,
    "preactresnet34": _resnet34_preact,
    "vgg16mod": _vgg16mod,
}


def _create_model(name: str, num_classes: int) -> torch.nn.Module:
    """Return a torch model architecture."""
    if name in _model_registry.keys():
        return _model_registry[name](num_classes)
    else:
        if name not in timm.list_models():
            raise UnknownArchitectureError(f"unknown architecture: '{name}'")
        return timm.create_model(name, num_classes=num_classes)


# Keys are tuple of (architecture, weights_name).
_known_model_weights: Dict[Tuple[str, str], Weights] = {}


def register_model_weights(root: Path):
    modeldefs = list(root.glob("*.yml")) + list(root.glob("*.yaml"))
    for modeldef in modeldefs:
        try:
            w = Weights.from_yaml(modeldef)
        except Exception as e:
            raise ModelRegistrationError(
                f"Error registering model from config file ('{modeldef}')\n"
                f"Original error is: {e}"
            )
        if w.architecture not in timm.list_models() + list(_model_registry.keys()):
            raise UnknownArchitectureError(f"{w.architecture} implementation not found")
        key = (w.architecture, w.name)
        if key in _known_model_weights:
            raise DuplicateModelWeights(
                f"duplicate models weights: {(w.architecture, w.name)}"
            )
        _known_model_weights[key] = w


def get_model_weights(architecture: str, name: str) -> Weights:
    """Get weights object for an architecture and weights name."""
    key = (architecture, name)
    try:
        return _known_model_weights[key]
    except KeyError:
        raise ModelWeightsNotFound(
            f"model weights are not found for architecture '{architecture}' and "
            f"weights name '{name}'. Available models are"
            f"{list_all_models_and_weights()}."
        )


register_model_weights(Path(__file__).parent / ".." / "modeldefs")

# Register any user-supplied configurations.
wsinfer_path = os.environ.get("WSINFER_PATH")
if wsinfer_path is not None:
    for path in wsinfer_path.split(":"):
        register_model_weights(Path(path))
    del path
del wsinfer_path


def list_all_models_and_weights() -> List[Tuple[str, str]]:
    """Return list of tuples of `(model_name, weights_name)` with available pairs."""
    vals = list(_known_model_weights.keys())
    vals.sort()
    return vals



import torch
import numpy as np
import torch.nn as nn
from torch.utils.data import (
    DataLoader,
    Sampler,
    WeightedRandomSampler,
    RandomSampler,
    SequentialSampler,
    sampler,
)
import torch.optim as optim
from itertools import islice
import math
import collections

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class SubsetSequentialSampler(Sampler):
    """Samples elements sequentially from a given list of indices, without replacement.

    Arguments:
            indices (sequence): a sequence of indices
    """

    def __init__(self, indices):
        self.indices = indices

    def __iter__(self):
        return iter(self.indices)

    def __len__(self):
        return len(self.indices)


def collate_MIL(batch):
    img = torch.cat([item[0] for item in batch], dim=0)
    label = torch.LongTensor([item[1] for item in batch])
    return [img, label]


def collate_features(batch):
    img = torch.cat([item[0] for item in batch], dim=0)
    coords = np.vstack([item[1] for item in batch])
    return [img, coords]


def get_simple_loader(dataset, batch_size=1, num_workers=1):
    kwargs = (
        {"pin_memory": False, "num_workers": num_workers}
        if device.type == "cuda"
        else {}
    )
    loader = DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=sampler.SequentialSampler(dataset),
        collate_fn=collate_MIL,
        **kwargs
    )
    return loader


def get_split_loader(split_dataset, training=False, testing=False, weighted=False):
    """
    return either the validation loader or training loader
    """
    kwargs = {"num_workers": 4} if device.type == "cuda" else {}
    if not testing:
        if training:
            if weighted:
                weights = make_weights_for_balanced_classes_split(split_dataset)
                loader = DataLoader(
                    split_dataset,
                    batch_size=1,
                    sampler=WeightedRandomSampler(weights, len(weights)),
                    collate_fn=collate_MIL,
                    **kwargs
                )
            else:
                loader = DataLoader(
                    split_dataset,
                    batch_size=1,
                    sampler=RandomSampler(split_dataset),
                    collate_fn=collate_MIL,
                    **kwargs
                )
        else:
            loader = DataLoader(
                split_dataset,
                batch_size=1,
                sampler=SequentialSampler(split_dataset),
                collate_fn=collate_MIL,
                **kwargs
            )

    else:
        ids = np.random.choice(
            np.arange(len(split_dataset), int(len(split_dataset) * 0.1)), replace=False
        )
        loader = DataLoader(
            split_dataset,
            batch_size=1,
            sampler=SubsetSequentialSampler(ids),
            collate_fn=collate_MIL,
            **kwargs
        )

    return loader


def get_optim(model, args):
    if args.opt == "adam":
        optimizer = optim.Adam(
            filter(lambda p: p.requires_grad, model.parameters()),
            lr=args.lr,
            weight_decay=args.reg,
        )
    elif args.opt == "sgd":
        optimizer = optim.SGD(
            filter(lambda p: p.requires_grad, model.parameters()),
            lr=args.lr,
            momentum=0.9,
            weight_decay=args.reg,
        )
    else:
        raise NotImplementedError
    return optimizer


def print_network(net):
    num_params = 0
    num_params_train = 0
    print(net)

    for param in net.parameters():
        n = param.numel()
        num_params += n
        if param.requires_grad:
            num_params_train += n

    print("Total number of parameters: %d" % num_params)
    print("Total number of trainable parameters: %d" % num_params_train)


def generate_split(
    cls_ids,
    val_num,
    test_num,
    samples,
    n_splits=5,
    seed=7,
    label_frac=1.0,
    custom_test_ids=None,
):
    indices = np.arange(samples).astype(int)

    if custom_test_ids is not None:
        indices = np.setdiff1d(indices, custom_test_ids)

    np.random.seed(seed)
    for i in range(n_splits):
        all_val_ids = []
        all_test_ids = []
        sampled_train_ids = []

        if custom_test_ids is not None:  # pre-built test split, do not need to sample
            all_test_ids.extend(custom_test_ids)

        for c in range(len(val_num)):
            possible_indices = np.intersect1d(
                cls_ids[c], indices
            )  # all indices of this class
            val_ids = np.random.choice(
                possible_indices, val_num[c], replace=False
            )  # validation ids

            remaining_ids = np.setdiff1d(
                possible_indices, val_ids
            )  # indices of this class left after validation
            all_val_ids.extend(val_ids)

            if custom_test_ids is None:  # sample test split

                test_ids = np.random.choice(remaining_ids, test_num[c], replace=False)
                remaining_ids = np.setdiff1d(remaining_ids, test_ids)
                all_test_ids.extend(test_ids)

            if label_frac == 1:
                sampled_train_ids.extend(remaining_ids)

            else:
                sample_num = math.ceil(len(remaining_ids) * label_frac)
                slice_ids = np.arange(sample_num)
                sampled_train_ids.extend(remaining_ids[slice_ids])

        yield sampled_train_ids, all_val_ids, all_test_ids


def nth(iterator, n, default=None):
    if n is None:
        return collections.deque(iterator, maxlen=0)
    else:
        return next(islice(iterator, n, None), default)


def calculate_error(Y_hat, Y):
    error = 1.0 - Y_hat.float().eq(Y.float()).float().mean().item()

    return error


def make_weights_for_balanced_classes_split(dataset):
    N = float(len(dataset))
    weight_per_class = [
        N / len(dataset.slide_cls_ids[c]) for c in range(len(dataset.slide_cls_ids))
    ]
    weight = [0] * int(N)
    for idx in range(len(dataset)):
        y = dataset.getlabel(idx)
        weight[idx] = weight_per_class[y]

    return torch.DoubleTensor(weight)


def initialize_weights(module):
    for m in module.modules():
        if isinstance(m, nn.Linear):
            nn.init.xavier_normal_(m.weight)
            m.bias.data.zero_()

        elif isinstance(m, nn.BatchNorm1d):
            nn.init.constant_(m.weight, 1)
            nn.init.constant_(m.bias, 0)



"""Convert CSV of model outputs to Stony Brook BMI formats.

Output directory tree for single class outputs:
├── heatmap_jsons
│   ├── heatmap-SLIDEID.json
│   └── meta-SLIDEID.json
└── heatmap_txt
    ├── color-SLIDEID
    └── prediction-SLIDEID

Output directory tree for multi-class outputs:
├── heatmap_jsons
│   └── CLASS_LABEL
│       ├── heatmap-SLIDEID.json
│       └── meta-SLIDEID.json
└── heatmap_txt
    └── CLASS_LABEL
        ├── color-SLIDEID
        └── prediction-SLIDEID
"""

import json
import multiprocessing
from pathlib import Path
import pprint
import random
import shutil
import time
import typing

import click
import large_image
import numpy as np
import pandas as pd
import tqdm


PathType = typing.Union[str, Path]


def _box_to_polygon(
    *, minx: float, miny: float, width: float, height: float
) -> typing.List[typing.Tuple[float, float]]:
    """Get coordinates of a box polygon."""
    maxx = minx + width
    maxy = miny + height
    return [(maxx, miny), (maxx, maxy), (minx, maxy), (minx, miny), (maxx, miny)]


def write_heatmap_and_meta_json_lines(
    input: PathType,
    output_heatmap: PathType,
    output_meta: PathType,
    slide_width: PathType,
    slide_height: PathType,
    execution_id: str,
    study_id: str,
    case_id: str,
    subject_id: str,
    class_name: str,
    run_metadata: typing.Dict,
) -> None:
    """Write JSON-lines files for one slide."""

    # Run this before defining the function so the entire JSON file has the same
    # execution time value.
    date = int(time.time())

    version_dict: typing.Dict = run_metadata["runtime"]["git"] or {}
    version_dict["model_path"] = run_metadata["model_weights"]["weights_file"]
    version_dict["model_hash"] = run_metadata["model_weights"]["weights_sha256"]
    version_dict["model_url"] = run_metadata["model_weights"]["weights_url"]
    version_dict["model_ver"] = None

    def row_to_json(row: pd.Series):
        minx, miny, width, height = (
            row["minx"],
            row["miny"],
            row["width"],
            row["height"],
        )
        patch_area_base_pixels = width * height
        # Scale the values to be a ratio of the whole slide dimensions. All of the
        # values in the dictionary (except 'footprint') use these normalized
        # coordinates.
        minx = float(minx / slide_width)
        miny = float(miny / slide_height)
        width = float(width / slide_width)
        height = float(height / slide_height)
        maxx = minx + width
        maxy = miny + height
        centerx = (minx + maxx) / 2
        centery = (miny + maxy) / 2
        coords = _box_to_polygon(minx=minx, miny=miny, width=width, height=height)

        # Get the probabilites from the model.
        if f"prob_{class_name}" not in row.index:
            raise KeyError(f"class name not found in results: {class_name}")
        class_probability: float = row[f"prob_{class_name}"]
        return {
            "type": "Feature",
            "parent_id": "self",
            "object_type": "heatmap_multiple",
            "x": centerx,
            "y": centery,
            "normalized": "true",
            "footprint": patch_area_base_pixels,
            "geometry": {
                "coordinates": [coords],
                "type": "Polygon",
            },
            "provenance": {
                "analysis": {
                    "source": "computer",
                    "execution_id": execution_id,
                    "cancer_type": "quip",
                    "study_id": study_id,
                    "computation": "heatmap",
                    "execution_time": run_metadata["timestamp"],
                },
                "image": {
                    "case_id": case_id,
                    "subject_id": subject_id,
                },
                "version": version_dict,
            },
            "bbox": [minx, miny, maxx, maxy],
            "properties": {
                "multiheat_param": {
                    "human_weight": -1,
                    "metric_array": [class_probability],
                    "heatname_array": [class_name],
                    "weight_array": ["1"],
                },
                "metric_value": class_probability,
                "metric_type": "tile_dice",
                "human_mark": -1,
            },
            "date": {"$date": date},
        }

    # Write heatmap JSON lines file.
    df = pd.read_csv(input)
    features = df.apply(row_to_json, axis=1)
    features = features.tolist()
    features = (json.dumps(row) for row in features)
    with open(output_heatmap, "w") as f:
        f.writelines(line + "\n" for line in features)

    # Write meta file.
    meta_dict = {
        "color": "yellow",  # This is copied from the lung cancer detection code.
        "title": execution_id,
        "image": {
            "case_id": case_id,
            "subject_id": subject_id,
        },
        "provenance": {
            "analysis_execution_id": execution_id,
            "analysis_execution_date": run_metadata["timestamp"],
            "study_id": study_id,
            "type": "computer",
            "version": version_dict,
        },
        "submit_date": {"$date": date},
        "randval": random.uniform(0, 1),
    }
    with open(output_meta, "w") as f:
        json.dump(meta_dict, f)


def write_heatmap_txt(input: PathType, output: PathType, class_names: typing.List[str]):
    df = pd.read_csv(input)
    # TODO: should we round and cast to int here?
    df.loc[:, "x_loc"] = (df.minx + (df.width / 2)).round().astype(int)
    df.loc[:, "y_loc"] = (df.miny + (df.height / 2)).round().astype(int)
    prob_cols = [f"prob_{c}" for c in class_names]
    cols = ["x_loc", "y_loc", *prob_cols]
    df = df.loc[:, cols]
    # Rename 'prob_NAME' columns to 'NAME'.
    df = df.rename(columns={c: c[5:] for c in prob_cols})
    df.to_csv(output, index=False, sep=" ")


def write_color_txt(
    input: PathType,
    output: PathType,
    ts: large_image.tilesource.TileSource,
    num_processes: int = 6,
):
    def whiteness(arr):
        arr = np.asarray(arr)
        return np.std(arr, axis=(0, 1)).mean()

    def blackness(arr):
        arr = np.asarray(arr)
        return arr.mean()

    def redness(arr):
        arr = np.asarray(arr)
        r, g, b = arr[:, :, 0], arr[:, :, 1], arr[:, :, 2]
        # boolean multiplication is logical and
        return np.mean((r >= 190) * (g <= 100) * (b <= 100))

    global get_color  # Hack to please multiprocessing.

    def get_color(row: pd.Series):
        arr, _ = ts.getRegion(
            format=large_image.constants.TILE_FORMAT_NUMPY,
            region=dict(
                left=row["minx"],
                top=row["miny"],
                width=row["width"],
                height=row["height"],
            ),
        )
        white = whiteness(arr)
        black = blackness(arr)
        red = redness(arr)
        return white, black, red

    df = pd.read_csv(input)
    df_rows_as_dicts = df.to_dict("records")

    # https://stackoverflow.com/a/45276885/5666087
    with multiprocessing.Pool(num_processes) as p:
        results = list(
            tqdm.tqdm(
                p.imap(get_color, df_rows_as_dicts, chunksize=16),  # type: ignore
                total=len(df_rows_as_dicts),
            )
        )
    df.loc[:, ["whiteness", "blackness", "redness"]] = results
    cx = df.minx + ((df.minx + df.width) / 2)
    assert np.array_equal(cx.astype(int), cx), "not all center x's are integer"
    df.loc[:, "cx"] = cx.astype(int)
    del cx  # sanity

    cy = df.miny + ((df.miny + df.height) / 2)
    assert np.array_equal(cy.astype(int), cy), "not all center y's are integer"
    df.loc[:, "cy"] = cy.astype(int)
    del cy

    df = df.loc[:, ["cx", "cy", "whiteness", "blackness", "redness"]]
    df.to_csv(output, header=False, index=False, sep=" ")


@click.command()
@click.argument(
    "results_dir",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=Path, resolve_path=True
    ),
)
@click.argument(
    "output",
    type=click.Path(exists=False, path_type=Path, resolve_path=True),
)
@click.option(
    "--wsi-dir",
    required=True,
    help="Directory with whole slide images.",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=Path, resolve_path=True
    ),
)
@click.option("--execution-id", required=True, help="Unique execution ID for this run.")
@click.option("--study-id", required=True, help="Study ID, like TCGA-BRCA.")
@click.option(
    "--make-color-text/--no-make-color-text",
    default=False,
    help="Make text files with color information for each patch. NOTE: this can add"
    " several minutes of processing time per slide.",
)
@click.option(
    "--num-processes",
    type=int,
    default=4,
    help="Number of processes to use when `--make-color-text` is enabled.",
)
def cli(
    *,
    results_dir: Path,
    output: Path,
    wsi_dir: Path,
    execution_id: str,
    study_id: str,
    make_color_text: bool = False,
    num_processes: int = 4,
):
    """Convert model outputs to Stony Brook format.

    Convert CSVs of patch predictions to .txt and .json formats for use with Stony
    Brook Biomedical Informatics viewers.

    RESULTS_DIR     Path to results directory (containing model-outputs dir).

    OUTPUT          Path to output directory in which to save files.
    """
    if not results_dir.exists():
        raise click.ClickException(f"results_dir does not exist: {results_dir}")
    if output.exists():
        raise click.ClickException("Output directory already exists.")
    if (
        not (results_dir / "model-outputs").exists()
        and (results_dir / "patches").exists()
    ):
        raise click.ClickException(
            "Model outputs have not been generated yet. Please run model inference."
        )
    if not (results_dir / "model-outputs").exists():
        raise click.ClickException(
            "Expected results_dir to contain a 'model-outputs' directory but it does"
            " not. Please provide the path to the directory that contains"
            " model-outputs, masks, and patches."
        )
    if not (results_dir / "run_metadata.json").exists():
        raise click.ClickException(
            f"Cannot find {results_dir / 'run_metadata.json'}. Was model inference run?"
        )
    if not wsi_dir.exists():
        raise click.ClickException("Whole slide image directory does not exist.")

    csvs = list((results_dir / "model-outputs").glob("*.csv"))
    if not csvs:
        raise click.ClickException("No CSVs found. Did you generate model outputs?")

    output.mkdir(exist_ok=False)

    with open(results_dir / "run_metadata.json") as f:
        run_metadata: typing.Dict = json.load(f)

    print("-" * 40)
    print("Run metadata")
    pprint.pprint(run_metadata, indent=2)
    print("-" * 40, "\n")

    # Get the output classes of the model. We will create separate directories for each
    # label name. But by default, we do not include labels that start with "no", like
    # "notils" and "notumor".
    class_names: typing.List[str] = run_metadata["model_weights"]["class_names"]
    ignore_names = {"notils", "notumor"}
    class_names = [n for n in class_names if n not in ignore_names]

    print("-" * 40)
    print("Output class names:")
    for c in class_names:
        print(f"  - {c}")
    print("-" * 40)

    for ii, input_csv in enumerate(csvs):
        print("-" * 40)
        print(f"Converting outputs for slide {ii+1} of {len(csvs)}")
        slide_id = input_csv.stem
        wsi_file = wsi_dir / f"{slide_id}.svs"
        if not wsi_file.exists():
            click.secho(f"WSI file not found: {wsi_file}", bg="red")
            click.secho("Skipping...", bg="red")
            continue
        ts = large_image.getTileSource(wsi_file)
        if ts.sizeX is None or ts.sizeY is None:
            click.secho(f"Unknown size for WSI: {wsi_file}", bg="red")
            click.secho("Skipping...", bg="red")
            continue

        for class_name in class_names:
            if len(class_names) == 1:
                # Do not use class-specific dirs if there is only one class.
                output_heatmap = output / "heatmap_json" / f"heatmap_{slide_id}.json"
                output_meta = output_heatmap.parent / f"meta_{slide_id}.json"
            else:
                output_heatmap = (
                    output / "heatmap_json" / class_name / f"heatmap_{slide_id}.json"
                )
            output_meta = output_heatmap.parent / f"meta_{slide_id}.json"
            output_heatmap.parent.mkdir(parents=True, exist_ok=True)

            click.echo("Writing JSON heatmap:")
            click.echo(f"  {output_heatmap}")
            click.echo("Writing JSON meta:")
            click.echo(f"  {output_meta}")
            write_heatmap_and_meta_json_lines(
                input=input_csv,
                output_heatmap=output_heatmap,
                output_meta=output_meta,
                slide_width=ts.sizeX,
                slide_height=ts.sizeY,
                execution_id=execution_id,
                study_id=study_id,
                case_id=slide_id,  # TODO: should case_id be different?
                subject_id=slide_id,
                class_name=class_name,
                run_metadata=run_metadata,
            )

            # Do not use class-specific dirs if there is only one class.
            if len(class_names) == 1:
                output_txt_prediction = (
                    output / "heatmap_txt" / f"prediction-{slide_id}"
                )
            else:
                output_txt_prediction = (
                    output / "heatmap_txt" / class_name / f"prediction-{slide_id}"
                )
            output_txt_prediction.parent.mkdir(parents=True, exist_ok=True)
            print("Writing TXT heatmap:")
            print(f"  {output_txt_prediction}")
            write_heatmap_txt(
                input=input_csv, output=output_txt_prediction, class_names=[class_name]
            )

        # Do this once per WSI. If the model has multi-class outputs, copy the color
        # file into each class-specific directory.
        if make_color_text:
            print("Writing color TXT file with color information for each patch")
            if len(class_names) == 1:
                output_color = output / "heatmap_txt" / f"color-{slide_id}"
                print(f"  {output_color}")
                write_color_txt(
                    input=input_csv,
                    output=output_color,
                    ts=ts,
                    num_processes=num_processes,
                )
            else:
                output_color = (
                    output / "heatmap_txt" / class_names[0] / f"color-{slide_id}"
                )
                print(f"  {output_color}")
                write_color_txt(
                    input=input_csv,
                    output=output_color,
                    ts=ts,
                    num_processes=num_processes,
                )
                # Copy this color file to all class-specific dirs.
                print("Copying color TXT file to class-specific TXT directories")
                for class_name in class_names[1:]:
                    cp_dst = output_color.parent.parent / class_name / output_color.name
                    print(f"  {cp_dst}")
                    shutil.copy(output_color, cp_dst)

    click.secho("Finished.", fg="green")



import click

# from .. import __version__
from .convert_csv_to_geojson import cli as _cli_convert_to_geojson
from .convert_csv_to_sbubmi import cli as _cli_convert_to_sbubmi
from .infer import cli as _cli_inference
from .list_models_and_weights import cli as _cli_list_models_weights
from .patch import cli as _cli_patch


# We use invoke_without_command=True so that 'wsinfer' on its own can be used for
# inference on slides.
@click.group()
@click.version_option()
def cli():
    """Run patch-level classification inference on whole slide images."""
    pass


cli.add_command(_cli_inference, name="run")
cli.add_command(_cli_convert_to_geojson, name="togeojson")
cli.add_command(_cli_convert_to_sbubmi, name="tosbu")
cli.add_command(_cli_patch, name="patch")
cli.add_command(_cli_list_models_weights, name="list")



import math
import os
from xml.dom import minidom
import multiprocessing as mp
import cv2
import numpy as np
import openslide
from PIL import Image

from .wsi_utils import (
    savePatchIter_bag_hdf5,
    initialize_hdf5_bag,
    save_hdf5,
    isBlackPatch,
    isWhitePatch,
)
from .util_classes import (
    isInContourV1,
    isInContourV2,
    isInContourV3_Easy,
    isInContourV3_Hard,
    Contour_Checking_fn,
)
from ..utils.file_utils import load_pkl, save_pkl

Image.MAX_IMAGE_PIXELS = 933120000


class WholeSlideImage(object):
    def __init__(self, path):

        """
        Args:
            path (str): fullpath to WSI file
        """

        #         self.name = ".".join(path.split("/")[-1].split('.')[:-1])
        self.name = os.path.splitext(os.path.basename(path))[0]
        self.wsi = openslide.open_slide(path)
        self.level_downsamples = self._assertLevelDownsamples()
        self.level_dim = self.wsi.level_dimensions

        self.contours_tissue = None
        self.contours_tumor = None
        self.hdf5_file = None

    def getOpenSlide(self):
        return self.wsi

    def initXML(self, xml_path):
        def _createContour(coord_list):
            return np.array(
                [
                    [
                        [
                            int(float(coord.attributes["X"].value)),
                            int(float(coord.attributes["Y"].value)),
                        ]
                    ]
                    for coord in coord_list
                ],
                dtype="int32",
            )

        xmldoc = minidom.parse(xml_path)
        annotations = [
            anno.getElementsByTagName("Coordinate")
            for anno in xmldoc.getElementsByTagName("Annotation")
        ]
        self.contours_tumor = [_createContour(coord_list) for coord_list in annotations]
        self.contours_tumor = sorted(
            self.contours_tumor, key=cv2.contourArea, reverse=True
        )

    def initTxt(self, annot_path):
        def _create_contours_from_dict(annot):
            all_cnts = []
            for idx, annot_group in enumerate(annot):
                contour_group = annot_group["coordinates"]
                if annot_group["type"] == "Polygon":
                    for idx, contour in enumerate(contour_group):
                        contour = np.array(contour).astype(np.int32).reshape(-1, 1, 2)
                        all_cnts.append(contour)

                else:
                    for idx, sgmt_group in enumerate(contour_group):
                        contour = []
                        for sgmt in sgmt_group:
                            contour.extend(sgmt)
                        contour = np.array(contour).astype(np.int32).reshape(-1, 1, 2)
                        all_cnts.append(contour)

            return all_cnts

        with open(annot_path, "r") as f:
            annot = f.read()
            annot = eval(annot)
        self.contours_tumor = _create_contours_from_dict(annot)
        self.contours_tumor = sorted(
            self.contours_tumor, key=cv2.contourArea, reverse=True
        )

    def initSegmentation(self, mask_file):
        # load segmentation results from pickle file
        asset_dict = load_pkl(mask_file)
        self.holes_tissue = asset_dict["holes"]
        self.contours_tissue = asset_dict["tissue"]

    def saveSegmentation(self, mask_file):
        # save segmentation results using pickle
        asset_dict = {"holes": self.holes_tissue, "tissue": self.contours_tissue}
        save_pkl(mask_file, asset_dict)

    def segmentTissue(
        self,
        seg_level=0,
        sthresh=20,
        sthresh_up=255,
        mthresh=7,
        close=0,
        use_otsu=False,
        filter_params={"a_t": 100},
        ref_patch_size=512,
        exclude_ids=[],
        keep_ids=[],
    ):
        """
        Segment the tissue via HSV -> Median thresholding -> Binary threshold
        """

        def _filter_contours(contours, hierarchy, filter_params):
            """
            Filter contours by: area.
            """
            filtered = []

            # find indices of foreground contours (parent == -1)
            hierarchy_1 = np.flatnonzero(hierarchy[:, 1] == -1)
            all_holes = []

            # loop through foreground contour indices
            for cont_idx in hierarchy_1:
                # actual contour
                cont = contours[cont_idx]
                # indices of holes contained in this contour
                # (children of parent contour)
                holes = np.flatnonzero(hierarchy[:, 1] == cont_idx)
                # take contour area (includes holes)
                a = cv2.contourArea(cont)
                # calculate the contour area of each hole
                hole_areas = [cv2.contourArea(contours[hole_idx]) for hole_idx in holes]
                # actual area of foreground contour region
                a = a - np.array(hole_areas).sum()
                if a == 0:
                    continue
                if tuple((filter_params["a_t"],)) < tuple((a,)):
                    filtered.append(cont_idx)
                    all_holes.append(holes)

            foreground_contours = [contours[cont_idx] for cont_idx in filtered]

            hole_contours = []

            for hole_ids in all_holes:
                unfiltered_holes = [contours[idx] for idx in hole_ids]
                unfilered_holes = sorted(
                    unfiltered_holes, key=cv2.contourArea, reverse=True
                )
                # take max_n_holes largest holes by area
                unfilered_holes = unfilered_holes[: filter_params["max_n_holes"]]
                filtered_holes = []

                # filter these holes
                for hole in unfilered_holes:
                    if cv2.contourArea(hole) > filter_params["a_h"]:
                        filtered_holes.append(hole)

                hole_contours.append(filtered_holes)

            return foreground_contours, hole_contours

        img = np.array(
            self.wsi.read_region((0, 0), seg_level, self.level_dim[seg_level])
        )
        img_hsv = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)  # Convert to HSV space
        img_med = cv2.medianBlur(img_hsv[:, :, 1], mthresh)  # Apply median blurring

        # Thresholding
        if use_otsu:
            _, img_otsu = cv2.threshold(
                img_med, 0, sthresh_up, cv2.THRESH_OTSU + cv2.THRESH_BINARY
            )
        else:
            _, img_otsu = cv2.threshold(img_med, sthresh, sthresh_up, cv2.THRESH_BINARY)

        # Morphological closing
        if close > 0:
            kernel = np.ones((close, close), np.uint8)
            img_otsu = cv2.morphologyEx(img_otsu, cv2.MORPH_CLOSE, kernel)

        scale = self.level_downsamples[seg_level]
        scaled_ref_patch_area = int(ref_patch_size**2 / (scale[0] * scale[1]))
        filter_params = filter_params.copy()
        filter_params["a_t"] = filter_params["a_t"] * scaled_ref_patch_area
        filter_params["a_h"] = filter_params["a_h"] * scaled_ref_patch_area

        # Find and filter contours
        contours, hierarchy = cv2.findContours(
            img_otsu, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE
        )  # Find contours
        hierarchy = np.squeeze(hierarchy, axis=(0,))[:, 2:]
        if filter_params:
            foreground_contours, hole_contours = _filter_contours(
                contours, hierarchy, filter_params
            )  # Necessary for filtering out artifacts

        self.contours_tissue = self.scaleContourDim(foreground_contours, scale)
        self.holes_tissue = self.scaleHolesDim(hole_contours, scale)

        # exclude_ids = [0,7,9]
        if len(keep_ids) > 0:
            contour_ids = set(keep_ids) - set(exclude_ids)
        else:
            contour_ids = set(np.arange(len(self.contours_tissue))) - set(exclude_ids)

        self.contours_tissue = [self.contours_tissue[i] for i in contour_ids]
        self.holes_tissue = [self.holes_tissue[i] for i in contour_ids]

    def visWSI(
        self,
        vis_level=0,
        color=(0, 255, 0),
        hole_color=(0, 0, 255),
        annot_color=(255, 0, 0),
        line_thickness=250,
        max_size=None,
        top_left=None,
        bot_right=None,
        custom_downsample=1,
        view_slide_only=False,
        number_contours=False,
        seg_display=True,
        annot_display=True,
    ):

        downsample = self.level_downsamples[vis_level]
        scale = [1 / downsample[0], 1 / downsample[1]]

        if top_left is not None and bot_right is not None:
            top_left = tuple(top_left)
            bot_right = tuple(bot_right)
            w, h = tuple(
                (np.array(bot_right) * scale).astype(int)
                - (np.array(top_left) * scale).astype(int)
            )
            region_size = (w, h)
        else:
            top_left = (0, 0)
            region_size = self.level_dim[vis_level]

        img = np.array(
            self.wsi.read_region(top_left, vis_level, region_size).convert("RGB")
        )

        if not view_slide_only:
            offset = tuple(-(np.array(top_left) * scale).astype(int))
            line_thickness = int(line_thickness * math.sqrt(scale[0] * scale[1]))
            if self.contours_tissue is not None and seg_display:
                if not number_contours:
                    cv2.drawContours(
                        img,
                        self.scaleContourDim(self.contours_tissue, scale),
                        -1,
                        color,
                        line_thickness,
                        lineType=cv2.LINE_8,
                        offset=offset,
                    )

                else:  # add numbering to each contour
                    for idx, cont in enumerate(self.contours_tissue):
                        contour = np.array(self.scaleContourDim(cont, scale))
                        M = cv2.moments(contour)
                        cX = int(M["m10"] / (M["m00"] + 1e-9))
                        cY = int(M["m01"] / (M["m00"] + 1e-9))
                        # draw the contour and put text next to center
                        cv2.drawContours(
                            img,
                            [contour],
                            -1,
                            color,
                            line_thickness,
                            lineType=cv2.LINE_8,
                            offset=offset,
                        )
                        cv2.putText(
                            img,
                            "{}".format(idx),
                            (cX, cY),
                            cv2.FONT_HERSHEY_SIMPLEX,
                            2,
                            (255, 0, 0),
                            10,
                        )

                for holes in self.holes_tissue:
                    cv2.drawContours(
                        img,
                        self.scaleContourDim(holes, scale),
                        -1,
                        hole_color,
                        line_thickness,
                        lineType=cv2.LINE_8,
                    )

            if self.contours_tumor is not None and annot_display:
                cv2.drawContours(
                    img,
                    self.scaleContourDim(self.contours_tumor, scale),
                    -1,
                    annot_color,
                    line_thickness,
                    lineType=cv2.LINE_8,
                    offset=offset,
                )

        img = Image.fromarray(img)

        w, h = img.size
        if custom_downsample > 1:
            img = img.resize((int(w / custom_downsample), int(h / custom_downsample)))

        if max_size is not None and (w > max_size or h > max_size):
            resizeFactor = max_size / w if w > h else max_size / h
            img = img.resize((int(w * resizeFactor), int(h * resizeFactor)))

        return img

    def createPatches_bag_hdf5(
        self,
        save_path,
        patch_level=0,
        patch_size=256,
        step_size=256,
        save_coord=True,
        **kwargs
    ):
        contours = self.contours_tissue

        print(
            "Creating patches for: ",
            self.name,
            "...",
        )
        for idx, cont in enumerate(contours):
            patch_gen = self._getPatchGenerator(
                cont, idx, patch_level, save_path, patch_size, step_size, **kwargs
            )

            if self.hdf5_file is None:
                try:
                    first_patch = next(patch_gen)

                # empty contour, continue
                except StopIteration:
                    continue

                file_path = initialize_hdf5_bag(first_patch, save_coord=save_coord)
                self.hdf5_file = file_path

            for patch in patch_gen:
                savePatchIter_bag_hdf5(patch)

        return self.hdf5_file

    def _getPatchGenerator(
        self,
        cont,
        cont_idx,
        patch_level,
        save_path,
        patch_size=256,
        step_size=256,
        custom_downsample=1,
        white_black=True,
        white_thresh=15,
        black_thresh=50,
        contour_fn="four_pt",
        use_padding=True,
    ):
        start_x, start_y, w, h = (
            cv2.boundingRect(cont)
            if cont is not None
            else (0, 0, self.level_dim[patch_level][0], self.level_dim[patch_level][1])
        )
        print("Bounding Box:", start_x, start_y, w, h)
        print("Contour Area:", cv2.contourArea(cont))

        if custom_downsample > 1:
            assert custom_downsample == 2
            target_patch_size = patch_size
            patch_size = target_patch_size * 2
            step_size = step_size * 2
            print(
                "Custom Downsample: {}, Patching at {} x {}, But Final Patch Size is"
                " {} x {}".format(
                    custom_downsample,
                    patch_size,
                    patch_size,
                    target_patch_size,
                    target_patch_size,
                )
            )

        patch_downsample = (
            int(self.level_downsamples[patch_level][0]),
            int(self.level_downsamples[patch_level][1]),
        )
        ref_patch_size = (
            patch_size * patch_downsample[0],
            patch_size * patch_downsample[1],
        )

        step_size_x = step_size * patch_downsample[0]
        step_size_y = step_size * patch_downsample[1]

        if isinstance(contour_fn, str):
            if contour_fn == "four_pt":
                cont_check_fn = isInContourV3_Easy(
                    contour=cont, patch_size=ref_patch_size[0], center_shift=0.5
                )
            elif contour_fn == "four_pt_hard":
                cont_check_fn = isInContourV3_Hard(
                    contour=cont, patch_size=ref_patch_size[0], center_shift=0.5
                )
            elif contour_fn == "center":
                cont_check_fn = isInContourV2(
                    contour=cont, patch_size=ref_patch_size[0]
                )
            elif contour_fn == "basic":
                cont_check_fn = isInContourV1(contour=cont)
            else:
                raise NotImplementedError
        else:
            assert isinstance(contour_fn, Contour_Checking_fn)
            cont_check_fn = contour_fn

        img_w, img_h = self.level_dim[0]
        if use_padding:
            stop_y = start_y + h
            stop_x = start_x + w
        else:
            stop_y = min(start_y + h, img_h - ref_patch_size[1])
            stop_x = min(start_x + w, img_w - ref_patch_size[0])

        count = 0
        for y in range(start_y, stop_y, step_size_y):
            for x in range(start_x, stop_x, step_size_x):

                if not self.isInContours(
                    cont_check_fn,
                    (x, y),
                    self.holes_tissue[cont_idx],
                    ref_patch_size[0],
                ):  # point not inside contour and its associated holes
                    continue

                count += 1
                patch_PIL = self.wsi.read_region(
                    (x, y), patch_level, (patch_size, patch_size)
                ).convert("RGB")
                if custom_downsample > 1:
                    patch_PIL = patch_PIL.resize((target_patch_size, target_patch_size))

                if white_black:
                    if isBlackPatch(
                        np.array(patch_PIL), rgbThresh=black_thresh
                    ) or isWhitePatch(np.array(patch_PIL), satThresh=white_thresh):
                        continue

                patch_info = {
                    "x": x // (patch_downsample[0] * custom_downsample),
                    "y": y // (patch_downsample[1] * custom_downsample),
                    "cont_idx": cont_idx,
                    "patch_level": patch_level,
                    "downsample": self.level_downsamples[patch_level],
                    "downsampled_level_dim": tuple(
                        np.array(self.level_dim[patch_level]) // custom_downsample
                    ),
                    "level_dim": self.level_dim[patch_level],
                    "patch_PIL": patch_PIL,
                    "name": self.name,
                    "save_path": save_path,
                }

                yield patch_info

        print("patches extracted: {}".format(count))

    @staticmethod
    def isInHoles(holes, pt, patch_size):
        for hole in holes:
            if (
                cv2.pointPolygonTest(
                    hole, (pt[0] + patch_size / 2, pt[1] + patch_size / 2), False
                )
                > 0
            ):
                return 1

        return 0

    @staticmethod
    def isInContours(cont_check_fn, pt, holes=None, patch_size=256):
        if cont_check_fn(pt):
            if holes is not None:
                return not WholeSlideImage.isInHoles(holes, pt, patch_size)
            else:
                return 1
        return 0

    @staticmethod
    def scaleContourDim(contours, scale):
        return [np.array(cont * scale, dtype="int32") for cont in contours]

    @staticmethod
    def scaleHolesDim(contours, scale):
        return [
            [np.array(hole * scale, dtype="int32") for hole in holes]
            for holes in contours
        ]

    def _assertLevelDownsamples(self):
        level_downsamples = []
        dim_0 = self.wsi.level_dimensions[0]

        for downsample, dim in zip(
            self.wsi.level_downsamples, self.wsi.level_dimensions
        ):
            estimated_downsample = (dim_0[0] / float(dim[0]), dim_0[1] / float(dim[1]))
            level_downsamples.append(estimated_downsample) if estimated_downsample != (
                downsample,
                downsample,
            ) else level_downsamples.append((downsample, downsample))

        return level_downsamples

    def process_contours(
        self, save_path, patch_level=0, patch_size=256, step_size=256, **kwargs
    ):
        save_path_hdf5 = os.path.join(save_path, str(self.name) + ".h5")
        print(
            "Creating patches for: ",
            self.name,
            "...",
        )
        n_contours = len(self.contours_tissue)
        print("Total number of contours to process: ", n_contours)
        fp_chunk_size = math.ceil(n_contours * 0.05)
        init = True
        for idx, cont in enumerate(self.contours_tissue):
            if (idx + 1) % fp_chunk_size == fp_chunk_size:
                print("Processing contour {}/{}".format(idx, n_contours))

            asset_dict, attr_dict = self.process_contour(
                cont,
                self.holes_tissue[idx],
                patch_level,
                save_path,
                patch_size,
                step_size,
                **kwargs
            )
            if len(asset_dict) > 0:
                if init:
                    save_hdf5(save_path_hdf5, asset_dict, attr_dict, mode="w")
                    init = False
                else:
                    save_hdf5(save_path_hdf5, asset_dict, mode="a")

        return self.hdf5_file

    def process_contour(
        self,
        cont,
        contour_holes,
        patch_level,
        save_path,
        patch_size=256,
        step_size=256,
        contour_fn="four_pt",
        use_padding=True,
        top_left=None,
        bot_right=None,
    ):
        start_x, start_y, w, h = (
            cv2.boundingRect(cont)
            if cont is not None
            else (0, 0, self.level_dim[patch_level][0], self.level_dim[patch_level][1])
        )

        patch_downsample = (
            int(self.level_downsamples[patch_level][0]),
            int(self.level_downsamples[patch_level][1]),
        )
        ref_patch_size = (
            patch_size * patch_downsample[0],
            patch_size * patch_downsample[1],
        )

        img_w, img_h = self.level_dim[0]
        if use_padding:
            stop_y = start_y + h
            stop_x = start_x + w
        else:
            stop_y = min(start_y + h, img_h - ref_patch_size[1] + 1)
            stop_x = min(start_x + w, img_w - ref_patch_size[0] + 1)

        print("Bounding Box:", start_x, start_y, w, h)
        print("Contour Area:", cv2.contourArea(cont))

        if bot_right is not None:
            stop_y = min(bot_right[1], stop_y)
            stop_x = min(bot_right[0], stop_x)
        if top_left is not None:
            start_y = max(top_left[1], start_y)
            start_x = max(top_left[0], start_x)

        if bot_right is not None or top_left is not None:
            w, h = stop_x - start_x, stop_y - start_y
            if w <= 0 or h <= 0:
                print("Contour is not in specified ROI, skip")
                return {}, {}
            else:
                print("Adjusted Bounding Box:", start_x, start_y, w, h)

        if isinstance(contour_fn, str):
            if contour_fn == "four_pt":
                cont_check_fn = isInContourV3_Easy(
                    contour=cont, patch_size=ref_patch_size[0], center_shift=0.5
                )
            elif contour_fn == "four_pt_hard":
                cont_check_fn = isInContourV3_Hard(
                    contour=cont, patch_size=ref_patch_size[0], center_shift=0.5
                )
            elif contour_fn == "center":
                cont_check_fn = isInContourV2(
                    contour=cont, patch_size=ref_patch_size[0]
                )
            elif contour_fn == "basic":
                cont_check_fn = isInContourV1(contour=cont)
            else:
                raise NotImplementedError
        else:
            assert isinstance(contour_fn, Contour_Checking_fn)
            cont_check_fn = contour_fn

        step_size_x = step_size * patch_downsample[0]
        step_size_y = step_size * patch_downsample[1]

        x_range = np.arange(start_x, stop_x, step=step_size_x)
        y_range = np.arange(start_y, stop_y, step=step_size_y)
        x_coords, y_coords = np.meshgrid(x_range, y_range, indexing="ij")
        coord_candidates = np.array(
            [x_coords.flatten(), y_coords.flatten()]
        ).transpose()

        num_workers = mp.cpu_count()
        if num_workers > 4:
            num_workers = 4
        pool = mp.Pool(num_workers)

        iterable = [
            (coord, contour_holes, ref_patch_size[0], cont_check_fn)
            for coord in coord_candidates
        ]
        results = pool.starmap(WholeSlideImage.process_coord_candidate, iterable)
        pool.close()
        results = np.array([result for result in results if result is not None])

        print("Extracted {} coordinates".format(len(results)))

        if len(results) > 1:
            asset_dict = {"coords": results}

            attr = {
                "patch_size": patch_size,  # To be considered...
                "patch_level": patch_level,
                "downsample": self.level_downsamples[patch_level],
                "downsampled_level_dim": tuple(np.array(self.level_dim[patch_level])),
                "level_dim": self.level_dim[patch_level],
                "name": self.name,
                "save_path": save_path,
            }

            attr_dict = {"coords": attr}
            return asset_dict, attr_dict

        else:
            return {}, {}

    @staticmethod
    def process_coord_candidate(coord, contour_holes, ref_patch_size, cont_check_fn):
        if WholeSlideImage.isInContours(
            cont_check_fn, coord, contour_holes, ref_patch_size
        ):
            return coord
        else:
            return None

    def block_blending(
        self,
        img,
        vis_level,
        top_left,
        bot_right,
        alpha=0.5,
        blank_canvas=False,
        block_size=1024,
    ):
        print("\ncomputing blend")
        downsample = self.level_downsamples[vis_level]
        w = img.shape[1]
        h = img.shape[0]
        block_size_x = min(block_size, w)
        block_size_y = min(block_size, h)
        print("using block size: {} x {}".format(block_size_x, block_size_y))

        shift = top_left  # amount shifted w.r.t. (0,0)
        for x_start in range(
            top_left[0], bot_right[0], block_size_x * int(downsample[0])
        ):
            for y_start in range(
                top_left[1], bot_right[1], block_size_y * int(downsample[1])
            ):
                # print(x_start, y_start)

                # 1. convert wsi coordinates to image coordinates via shift and scale
                x_start_img = int((x_start - shift[0]) / int(downsample[0]))
                y_start_img = int((y_start - shift[1]) / int(downsample[1]))

                # 2. compute end points of blend tile, careful not to go over the edge
                # of the image
                y_end_img = min(h, y_start_img + block_size_y)
                x_end_img = min(w, x_start_img + block_size_x)

                if y_end_img == y_start_img or x_end_img == x_start_img:
                    continue
                # print('start_coord: {} end_coord: {}'.format(
                # (x_start_img, y_start_img), (x_end_img, y_end_img)))

                # 3. fetch blend block and size
                blend_block = img[y_start_img:y_end_img, x_start_img:x_end_img]
                blend_block_size = (x_end_img - x_start_img, y_end_img - y_start_img)

                if not blank_canvas:
                    # 4. read actual wsi block as canvas block
                    pt = (x_start, y_start)
                    canvas = np.array(
                        self.wsi.read_region(pt, vis_level, blend_block_size).convert(
                            "RGB"
                        )
                    )
                else:
                    # 4. OR create blank canvas block
                    canvas = np.array(
                        Image.new(
                            size=blend_block_size, mode="RGB", color=(255, 255, 255)
                        )
                    )

                # 5. blend color block and canvas block
                img[y_start_img:y_end_img, x_start_img:x_end_img] = cv2.addWeighted(
                    blend_block, alpha, canvas, 1 - alpha, 0, canvas
                )
        return img

    def get_seg_mask(self, region_size, scale, use_holes=False, offset=(0, 0)):
        print("\ncomputing foreground tissue mask")
        tissue_mask = np.full(np.flip(region_size), 0).astype(np.uint8)
        contours_tissue = self.scaleContourDim(self.contours_tissue, scale)
        offset = tuple((np.array(offset) * np.array(scale) * -1).astype(np.int32))

        contours_holes = self.scaleHolesDim(self.holes_tissue, scale)
        contours_tissue, contours_holes = zip(
            *sorted(
                zip(contours_tissue, contours_holes),
                key=lambda x: cv2.contourArea(x[0]),
                reverse=True,
            )
        )
        for idx in range(len(contours_tissue)):
            cv2.drawContours(
                image=tissue_mask,
                contours=contours_tissue,
                contourIdx=idx,
                color=(1),
                offset=offset,
                thickness=-1,
            )

            if use_holes:
                cv2.drawContours(
                    image=tissue_mask,
                    contours=contours_holes[idx],
                    contourIdx=-1,
                    color=(0),
                    offset=offset,
                    thickness=-1,
                )
            # contours_holes = self._scaleContourDim(self.holes_tissue, scale,
            # holes=True, area_thresh=area_thresh)

        tissue_mask = tissue_mask.astype(bool)
        print(
            "detected {}/{} of region as tissue".format(
                tissue_mask.sum(), tissue_mask.size
            )
        )
        return tissue_mask



from typing import Optional

import click

from .._patchlib.create_patches_fp import create_patches as _create_patches


@click.command()
@click.option(
    "--source",
    required=True,
    type=click.Path(exists=True),
    help="patch to directory containing whole slide image files",
)
@click.option("--step-size", default=None, type=int, help="Step size for patching.")
@click.option("--patch-size", required=True, type=int, help="Patch size in pixels.")
@click.option("--no-auto-skip", is_flag=True)
@click.option(
    "--save-dir",
    required=True,
    type=click.Path(),
    help="Directory to save processed data",
)
@click.option(
    "--preset",
    default="tcga.csv",
    help="Predefined profile of default segmentation and filter parameters (.csv)",
)
@click.option(
    "--process-list", help="Name of list of images to process with parameters (.csv)"
)
@click.option(
    "--patch-spacing",
    required=True,
    type=float,
    help="Patch spacing in micrometers per pixel.",
)
def cli(
    source: str,
    step_size: Optional[int],
    patch_size: int,
    no_auto_skip: bool,
    save_dir: str,
    preset: Optional[str],
    process_list: Optional[str],
    patch_spacing: float,
):
    """Patchify a directory of whole slide images."""

    print("create_patches_fp.py  Copyright (C) 2022  Mahmood Lab")
    print("This program comes with ABSOLUTELY NO WARRANTY.")
    print("This is free software, and you are welcome to redistribute it")
    print("under certain conditions.")

    _create_patches(
        source=source,
        step_size=step_size,
        patch_size=patch_size,
        patch_spacing=patch_spacing,
        save_dir=save_dir,
        patch=True,
        seg=True,
        stitch=True,
        no_auto_skip=no_auto_skip,
        preset=preset,
        process_list=process_list,
    )



"""PyTorch image classification transform.

From
https://github.com/pytorch/vision/blob/528651a031a08f9f97cc75bd619a326387708219/torchvision/transforms/_presets.py#L1
"""

from typing import Tuple
from typing import Union

from PIL import Image
import torch
from torchvision.transforms import functional as F

# Get the interpolation mode while catering to older (and newer) versions of
# torchvision and PIL.
if hasattr(F, "InterpolationMode"):
    BICUBIC = F.InterpolationMode.BICUBIC
    BILINEAR = F.InterpolationMode.BILINEAR
    LINEAR = F.InterpolationMode.BILINEAR
    NEAREST = F.InterpolationMode.NEAREST
elif hasattr(Image, "Resampling"):
    BICUBIC = Image.Resampling.BICUBIC
    BILINEAR = Image.Resampling.BILINEAR
    LINEAR = Image.Resampling.BILINEAR
    NEAREST = Image.Resampling.NEAREST
else:
    BICUBIC = Image.BICUBIC
    BILINEAR = Image.BILINEAR
    NEAREST = Image.NEAREST
    LINEAR = Image.LINEAR


class PatchClassification(torch.nn.Module):
    """Transform module to process RGB patches."""

    def __init__(
        self,
        *,
        resize_size: int,
        mean: Tuple[float, float, float],
        std: Tuple[float, float, float],
        interpolation=BILINEAR,
    ) -> None:
        super().__init__()
        self.resize_size = resize_size
        self.mean = list(mean)
        self.std = list(std)
        self.interpolation = interpolation

    def __repr__(self):
        return (
            f"PatchClassification(resize_size={self.resize_size}, mean={self.mean},"
            f" std={self.std}, interpolation={self.interpolation})"
        )

    def forward(self, input: Union[Image.Image, torch.Tensor]) -> torch.Tensor:
        img = F.resize(
            input,
            [self.resize_size, self.resize_size],
            interpolation=self.interpolation,
        )
        if not isinstance(img, torch.Tensor):
            img = F.pil_to_tensor(img)
        img = F.convert_image_dtype(img, torch.float)
        img = F.normalize(img, mean=self.mean, std=self.std)
        return img



"""Detect cancerous regions in a whole slide image."""

import getpass
import json
import os
from pathlib import Path
import platform
import subprocess
import sys
import typing

import click

from .._modellib.run_inference import run_inference
from .._modellib import models
from .._patchlib.create_dense_patch_grid import create_grid_and_save_multi_slides
from .._patchlib.create_patches_fp import create_patches

PathType = typing.Union[str, Path]


def _inside_container() -> str:
    if Path("/.dockerenv").exists():
        return "yes, docker"
    elif (
        Path("/singularity").exists()
        or Path("/singularity.d").exists()
        or Path("/.singularity.d").exists()
    ):
        # TODO: apptainer might change the name of this directory.
        return "yes, apptainer/singularity"
    return "no"


def _get_timestamp() -> str:
    from datetime import datetime

    dt = datetime.now().astimezone()
    # Thu Aug 25 23:32:17 2022 EDT
    return dt.strftime("%c %Z")


def _print_system_info() -> None:
    """Print information about the system."""
    import torch
    import torchvision
    from .. import __version__

    click.secho(f"\nRunning wsinfer version {__version__}", fg="green")
    print("\nIf you run into issues, please submit a new issue at")
    print("https://github.com/kaczmarj/patch-classification-pipeline/issues/new")
    print("\nSystem information")
    print("------------------")
    print(f"Timestamp: {_get_timestamp()}")
    print(f"{platform.platform()}")
    try:
        print(f"User: {getpass.getuser()}")
    except KeyError:
        # If /etc/passwd does not include the username of the current user ID, a
        # KeyError is thrown. This could happen in a Docker image running as a different
        # user with `--user $(id -u):$(id -g)` but that does not bind mount the
        # /etc/passwd file.
        print("User: UNKNOWN")
    print(f"Hostname: {platform.node()}")
    print(f"Working directory: {os.getcwd()}")
    print(f"In container: {_inside_container()}")
    print(f"Python executable: {sys.executable}")
    print(f"Python version: {platform.python_version()}")
    print(f"  Torch version: {torch.__version__}")
    print(f"  Torchvision version: {torchvision.__version__}")
    cuda_is_available = torch.cuda.is_available()
    if cuda_is_available:
        click.secho("GPU available", fg="green")
        cuda_ver = torch.version.cuda or "NOT FOUND"
        print(f"  CUDA version: {cuda_ver}")
    else:
        click.secho("GPU not available", bg="red", fg="black")
    cuda_visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", "NOT SET")
    print(f"CUDA_VISIBLE_DEVICES: {cuda_visible_devices}")
    if torch.version.cuda is None:
        click.secho("\n*******************************************", fg="yellow")
        click.secho("GPU WILL NOT BE USED", fg="yellow")
        if torch.version.cuda is None:
            click.secho("  CPU-only version of PyTorch is installed", fg="yellow")
        click.secho("*******************************************", fg="yellow")
    elif not torch.cuda.is_available():
        click.secho("\n*******************************************", fg="yellow")
        click.secho("GPU WILL NOT BE USED", fg="yellow")
        if torch.version.cuda is None:
            click.secho("  CUDA DEVICES NOT AVAILABLE", fg="yellow")
        click.secho("*******************************************", fg="yellow")


def _get_info_for_save(weights: models.Weights):
    """Get dictionary with information about the run. To save as JSON in output dir."""

    import torch
    from .. import __version__

    here = Path(__file__).parent.resolve()

    def get_git_info():
        here = Path(__file__).parent.resolve()

        def get_stdout(args) -> str:
            proc = subprocess.run(args, capture_output=True, cwd=here)
            return "" if proc.returncode != 0 else proc.stdout.decode().strip()

        git_remote = get_stdout("git config --get remote.origin.url".split())
        git_branch = get_stdout("git rev-parse --abbrev-ref HEAD".split())
        git_commit = get_stdout("git rev-parse HEAD".split())

        # https://stackoverflow.com/a/3879077/5666087
        cmd = subprocess.run("git diff-index --quiet HEAD --".split(), cwd=here)
        uncommitted_changes = cmd.returncode != 0
        return {
            "git_remote": git_remote,
            "git_branch": git_branch,
            "git_commit": git_commit,
            "uncommitted_changes": uncommitted_changes,
        }

    # Test if we are in a git repo. If we are, then get git info.
    cmd = subprocess.run(
        "git branch".split(),
        cwd=here,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    if cmd.returncode == 0:
        git_info = get_git_info()
    else:
        git_info = None
    del cmd, here  # For sanity.

    weights_file = weights.file
    if weights_file is None:
        if weights.url_file_name is None:
            raise TypeError("url_file_name must not be None if file is None.")
        weights_file = str(
            Path(torch.hub.get_dir()) / "checkpoints" / weights.url_file_name
        )

    return {
        "model_weights": {
            "name": weights.name,
            "architecture": weights.architecture,
            "weights_url": weights.url,
            "weights_url_file_name": weights.url_file_name,
            "weights_file": weights_file,
            "weights_sha256": weights.get_sha256_of_weights(),
            "class_names": weights.class_names,
            "num_classes": weights.num_classes,
            "patch_size_pixels": weights.patch_size_pixels,
            "spacing_um_px": weights.spacing_um_px,
            "transform": {
                "resize_size": weights.transform.resize_size,
                "mean": weights.transform.mean,
                "std": weights.transform.std,
            },
            "metadata": weights.metadata or None,
        },
        "runtime": {
            "version": __version__,
            "working_dir": os.getcwd(),
            "args": " ".join(sys.argv),
            "python_executable": sys.executable,
            "python_version": platform.python_version(),
            "in_container": _inside_container(),
            "pytorch_version": torch.__version__,
            "cuda_version": torch.version.cuda,
            "git": git_info,
        },
        "timestamp": _get_timestamp(),
    }


@click.command(context_settings=dict(auto_envvar_prefix="WSINFER"))
@click.pass_context
@click.option(
    "--wsi-dir",
    type=click.Path(exists=True, file_okay=False, path_type=Path, resolve_path=True),
    required=True,
    help="Directory containing whole slide images. This directory can *only* contain"
    " whole slide images.",
)
@click.option(
    "--results-dir",
    type=click.Path(file_okay=False, path_type=Path, resolve_path=True),
    required=True,
    help="Directory to store results. If directory exists, will skip"
    " whole slides for which outputs exist.",
)
@click.option(
    "--model",
    type=click.Choice([arch for arch, _ in models.list_all_models_and_weights()]),
    help="Model architecture to use. Not required if 'config' is used.",
)
@click.option(
    "--weights",
    type=click.Choice([w for _, w in models.list_all_models_and_weights()]),
    help="Name of weights to use for the model. Not required if 'config' is used.",
)
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False, path_type=Path, resolve_path=True),
    help=(
        "Path to configuration for architecture and weights. Use this option if the"
        " model weights are not registered in wsinfer. Mutually exclusive with"
        " 'model' and 'weights'."
    ),
)
@click.option(
    "--batch-size",
    type=click.IntRange(min=1),
    default=32,
    show_default=True,
    help="Batch size during model inference.",
)
@click.option(
    "--num-workers",
    default=0,
    show_default=True,
    type=click.IntRange(min=0),
    help="Number of workers to use for data loading during model inference (default=0"
    " for single thread). A reasonable value is 8.",
)
@click.option(
    "--dense-grid/--no-dense-grid",
    default=False,
    show_default=True,
    help="Use a dense grid of patch coordinates. Patches will be present even if no"
    " tissue is present",
)
def cli(
    ctx: click.Context,
    *,
    wsi_dir: Path,
    results_dir: Path,
    model: typing.Optional[str],
    weights: typing.Optional[str],
    config: typing.Optional[Path],
    batch_size: int,
    num_workers: int = 0,
    dense_grid: bool = False,
):
    """Run model inference on a directory of whole slide images.

    This command will create a tissue mask of each WSI. Then patch coordinates will be
    computed. The chosen model will be applied to each patch, and the results will be
    saved to a CSV in `RESULTS_DIR/model-output`.

    Example:

    CUDA_VISIBLE_DEVICES=0 wsinfer run --wsi_dir slides/ --results_dir results
    --model resnet34 --weights TCGA-BRCA-v1 --batch_size 32 --num_workers 4

    To list all available models and weights, use `wsinfer list`.
    """
    if model is None and weights is None and config is None:
        raise click.UsageError("one of (model and weights) or config is required.")
    elif (model is not None or weights is not None) and config is not None:
        raise click.UsageError("model and weights are mutually exclusive with config.")
    elif (model is not None) ^ (weights is not None):  # XOR
        raise click.UsageError("model and weights must both be set if one is set.")

    wsi_dir = wsi_dir.resolve()
    results_dir = results_dir.resolve()

    if not wsi_dir.exists():
        raise FileNotFoundError(f"Whole slide image directory not found: {wsi_dir}")

    # Test that wsi dir actually includes files. This is here for an interesting edge
    # case. When using a Linux container and if the data directory is symlinked from a
    # different directory, both directories need to be bind mounted onto the container.
    # If only the symlinked directory is included, then the patching script will fail,
    # even though it looks like there are files in the wsi_dir directory.
    files_in_wsi_dir = [p for p in wsi_dir.glob("*") if p.exists()]
    if not files_in_wsi_dir:
        raise FileNotFoundError(f"no files exist in the slide directory: {wsi_dir}")

    _print_system_info()

    print("\nCommand line arguments")
    print("----------------------")
    for key, value in ctx.params.items():
        print(f"{key} = {value}")
    print("----------------------\n")

    # Get weights object before running the patching script because we need to get the
    # necessary spacing and patch size.
    if model is not None and weights is not None:
        weights_obj = models.get_model_weights(model, name=weights)
    elif config is not None:
        weights_obj = models.Weights.from_yaml(config)

    click.secho("\nFinding patch coordinates...\n", fg="green")
    if dense_grid:
        click.echo("Not using a tissue mask.")
        create_grid_and_save_multi_slides(
            wsi_dir=wsi_dir,
            results_dir=results_dir,
            orig_patch_size=weights_obj.patch_size_pixels,
            patch_spacing_um_px=weights_obj.spacing_um_px,
        )
    else:
        create_patches(
            source=str(wsi_dir),
            save_dir=str(results_dir),
            patch_size=weights_obj.patch_size_pixels,
            patch_spacing=weights_obj.spacing_um_px,
            seg=True,
            patch=True,
            preset="tcga.csv",
        )

    click.secho("\nRunning model inference.\n", fg="green")
    run_inference(
        wsi_dir=wsi_dir,
        results_dir=results_dir,
        weights=weights_obj,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    run_metadata_outpath = results_dir / "run_metadata.json"
    click.echo(f"Saving metadata about run to {run_metadata_outpath}")
    run_metadata = _get_info_for_save(weights_obj)
    with open(run_metadata_outpath, "w") as f:
        json.dump(run_metadata, f, indent=2)

    click.secho("Finished.", fg="green")



import pickle
import h5py


def save_pkl(filename, save_object):
    writer = open(filename, "wb")
    pickle.dump(save_object, writer)
    writer.close()


def load_pkl(filename):
    loader = open(filename, "rb")
    file = pickle.load(loader)
    loader.close()
    return file


def save_hdf5(output_path, asset_dict, attr_dict=None, mode="a"):
    file = h5py.File(output_path, mode)
    for key, val in asset_dict.items():
        data_shape = val.shape
        if key not in file:
            data_type = val.dtype
            chunk_shape = (1,) + data_shape[1:]
            maxshape = (None,) + data_shape[1:]
            dset = file.create_dataset(
                key,
                shape=data_shape,
                maxshape=maxshape,
                chunks=chunk_shape,
                dtype=data_type,
            )
            dset[:] = val
            if attr_dict is not None:
                if key in attr_dict.keys():
                    for attr_key, attr_val in attr_dict[key].items():
                        dset.attrs[attr_key] = attr_val
        else:
            dset = file[key]
            dset.resize(len(dset) + data_shape[0], axis=0)
            dset[-data_shape[0] :] = val
    file.close()
    return output_path



# Create tissue mask and patch a whole slide image.
# Copyright (C) 2022  Mahmood Lab
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Modified by Jakub Kaczmarzyk (@kaczmarj on GitHub)
# - add --patch_spacing command line arg to request a patch size at a particular
#   spacing. The patch coordinates are calculated at the base (highest) resolution.
# - format code with black



# Create tissue mask and patch a whole slide image.
# Copyright (C) 2022  Mahmood Lab
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Modified by Jakub Kaczmarzyk (@kaczmarj on GitHub)
# - add --patch_spacing command line arg to request a patch size at a particular
#   spacing. The patch coordinates are calculated at the base (highest) resolution.
# - format code with black

"""Create tissue mask and patch a whole slide image.
Copyright (C) 2022  Mahmood Lab

Modified by Jakub Kaczmarzyk.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
"""

# internal imports
from .wsi_core.WholeSlideImage import WholeSlideImage
from .wsi_core.wsi_utils import StitchCoords
from .wsi_core.batch_process_utils import initialize_df

# other imports
import os
import pathlib
import numpy as np
import time
import pandas as pd
from typing import Optional

_script_path = pathlib.Path(__file__).resolve().parent


def stitching(file_path, wsi_object, downscale=64):
    start = time.time()
    heatmap = StitchCoords(
        file_path,
        wsi_object,
        downscale=downscale,
        bg_color=(0, 0, 0),
        alpha=-1,
        draw_grid=False,
    )
    total_time = time.time() - start

    return heatmap, total_time


def segment(WSI_object, seg_params=None, filter_params=None, mask_file=None):
    # Start Seg Timer
    start_time = time.time()
    # Use segmentation file
    if mask_file is not None:
        WSI_object.initSegmentation(mask_file)
    # Segment
    else:
        WSI_object.segmentTissue(**seg_params, filter_params=filter_params)

    # Stop Seg Timers
    seg_time_elapsed = time.time() - start_time
    return WSI_object, seg_time_elapsed


def patching(WSI_object, **kwargs):
    # Start Patch Timer
    start_time = time.time()

    # Patch
    file_path = WSI_object.process_contours(**kwargs)

    # Stop Patch Timer
    patch_time_elapsed = time.time() - start_time
    return file_path, patch_time_elapsed


def seg_and_patch(
    source,
    save_dir,
    patch_save_dir,
    mask_save_dir,
    stitch_save_dir,
    patch_size=256,
    step_size=256,
    seg_params={
        "seg_level": -1,
        "sthresh": 8,
        "mthresh": 7,
        "close": 4,
        "use_otsu": False,
        "keep_ids": "none",
        "exclude_ids": "none",
    },
    filter_params={"a_t": 100, "a_h": 16, "max_n_holes": 8},
    vis_params={"vis_level": -1, "line_thickness": 500},
    patch_params={"use_padding": True, "contour_fn": "four_pt"},
    patch_level=0,
    use_default_params=False,
    seg=False,
    save_mask=True,
    stitch=False,
    patch=False,
    auto_skip=True,
    process_list=None,
    patch_spacing=None,
):

    slides = sorted(os.listdir(source))
    slides = [slide for slide in slides if os.path.isfile(os.path.join(source, slide))]
    if process_list is None:
        df = initialize_df(slides, seg_params, filter_params, vis_params, patch_params)

    else:
        df = pd.read_csv(process_list)
        df = initialize_df(df, seg_params, filter_params, vis_params, patch_params)

    mask = df["process"] == 1
    process_stack = df[mask]

    total = len(process_stack)

    legacy_support = "a" in df.keys()
    if legacy_support:
        print("detected legacy segmentation csv file, legacy support enabled")
        df = df.assign(
            **{
                "a_t": np.full((len(df)), int(filter_params["a_t"]), dtype=np.uint32),
                "a_h": np.full((len(df)), int(filter_params["a_h"]), dtype=np.uint32),
                "max_n_holes": np.full(
                    (len(df)), int(filter_params["max_n_holes"]), dtype=np.uint32
                ),
                "line_thickness": np.full(
                    (len(df)), int(vis_params["line_thickness"]), dtype=np.uint32
                ),
                "contour_fn": np.full((len(df)), patch_params["contour_fn"]),
            }
        )

    seg_times = 0.0
    patch_times = 0.0
    stitch_times = 0.0

    orig_patch_size = patch_size

    for i in range(total):
        df.to_csv(os.path.join(save_dir, "process_list_autogen.csv"), index=False)
        idx = process_stack.index[i]
        slide = process_stack.loc[idx, "slide_id"]
        print("\n\nprogress: {:.2f}, {}/{}".format(i / total, i, total))
        print("processing {}".format(slide))

        df.loc[idx, "process"] = 0
        slide_id, _ = os.path.splitext(slide)

        if auto_skip and os.path.isfile(os.path.join(patch_save_dir, slide_id + ".h5")):
            print("{} already exist in destination location, skipped".format(slide_id))
            df.loc[idx, "status"] = "already_exist"
            continue

        # Inialize WSI
        full_path = os.path.join(source, slide)
        WSI_object = WholeSlideImage(full_path)

        if use_default_params:
            current_vis_params = vis_params.copy()
            current_filter_params = filter_params.copy()
            current_seg_params = seg_params.copy()
            current_patch_params = patch_params.copy()

        else:
            current_vis_params = {}
            current_filter_params = {}
            current_seg_params = {}
            current_patch_params = {}

            for key in vis_params.keys():
                if legacy_support and key == "vis_level":
                    df.loc[idx, key] = -1
                current_vis_params.update({key: df.loc[idx, key]})

            for key in filter_params.keys():
                if legacy_support and key == "a_t":
                    old_area = df.loc[idx, "a"]
                    seg_level = df.loc[idx, "seg_level"]
                    scale = WSI_object.level_downsamples[seg_level]
                    adjusted_area = int(old_area * (scale[0] * scale[1]) / (512 * 512))
                    current_filter_params.update({key: adjusted_area})
                    df.loc[idx, key] = adjusted_area
                current_filter_params.update({key: df.loc[idx, key]})

            for key in seg_params.keys():
                if legacy_support and key == "seg_level":
                    df.loc[idx, key] = -1
                current_seg_params.update({key: df.loc[idx, key]})

            for key in patch_params.keys():
                current_patch_params.update({key: df.loc[idx, key]})

        if current_vis_params["vis_level"] < 0:
            if len(WSI_object.level_dim) == 1:
                current_vis_params["vis_level"] = 0

            else:
                wsi = WSI_object.getOpenSlide()
                best_level = wsi.get_best_level_for_downsample(64)
                current_vis_params["vis_level"] = best_level

        if current_seg_params["seg_level"] < 0:
            if len(WSI_object.level_dim) == 1:
                current_seg_params["seg_level"] = 0

            else:
                wsi = WSI_object.getOpenSlide()
                best_level = wsi.get_best_level_for_downsample(64)
                current_seg_params["seg_level"] = best_level

        keep_ids = str(current_seg_params["keep_ids"])
        if keep_ids != "none" and len(keep_ids) > 0:
            str_ids = current_seg_params["keep_ids"]
            current_seg_params["keep_ids"] = np.array(str_ids.split(",")).astype(int)
        else:
            current_seg_params["keep_ids"] = []

        exclude_ids = str(current_seg_params["exclude_ids"])
        if exclude_ids != "none" and len(exclude_ids) > 0:
            str_ids = current_seg_params["exclude_ids"]
            current_seg_params["exclude_ids"] = np.array(str_ids.split(",")).astype(int)
        else:
            current_seg_params["exclude_ids"] = []

        w, h = WSI_object.level_dim[current_seg_params["seg_level"]]
        if w * h > 1e8:
            print(
                "level_dim {} x {} is likely too large for successful segmentation,"
                " aborting".format(w, h)
            )
            df.loc[idx, "status"] = "failed_seg"
            continue

        df.loc[idx, "vis_level"] = current_vis_params["vis_level"]
        df.loc[idx, "seg_level"] = current_seg_params["seg_level"]

        seg_time_elapsed = -1
        if seg:
            WSI_object, seg_time_elapsed = segment(
                WSI_object, current_seg_params, current_filter_params
            )

        if save_mask:
            mask = WSI_object.visWSI(**current_vis_params)
            mask_path = os.path.join(mask_save_dir, slide_id + ".jpg")
            mask.save(mask_path)

        patch_time_elapsed = -1  # Default time
        if patch:
            # -----------------------------------------------------------------------
            # Added by Jakub Kaczmarzyk (github kaczmarj) to get patch size for a
            # particular spacing. The patching happens at the highest resolution, but
            # we want to extract patches at a particular spacing.
            if patch_spacing is not None:
                from PIL import Image

                orig_max = Image.MAX_IMAGE_PIXELS
                import large_image

                # Importing large_image changes MAX_IMAGE_PIXELS to None.
                Image.MAX_IMAGE_PIXELS = orig_max
                del orig_max, Image

                ts = large_image.getTileSource(full_path)
                if ts.getMetadata()["mm_x"] is None:
                    print("!" * 40)
                    print("SKIPPING this slide because I cannot find the spacing!")
                    print(full_path)
                    print("!" * 40)
                    continue
                patch_mm = patch_spacing / 1000  # convert micrometer to millimeter.
                patch_size = orig_patch_size * patch_mm / ts.getMetadata()["mm_x"]
                patch_size = round(patch_size)
                del ts
            # Use non-overlapping patches by default.
            step_size = step_size or patch_size
            # ----------------------------------------------------------------------

            current_patch_params.update(
                {
                    "patch_level": patch_level,
                    "patch_size": patch_size,
                    "step_size": step_size,
                    "save_path": patch_save_dir,
                }
            )
            file_path, patch_time_elapsed = patching(
                WSI_object=WSI_object,
                **current_patch_params,
            )

        stitch_time_elapsed = -1
        if stitch:
            file_path = os.path.join(patch_save_dir, slide_id + ".h5")
            if os.path.isfile(file_path):
                heatmap, stitch_time_elapsed = stitching(
                    file_path, WSI_object, downscale=64
                )
                stitch_path = os.path.join(stitch_save_dir, slide_id + ".jpg")
                heatmap.save(stitch_path)

        print("segmentation took {} seconds".format(seg_time_elapsed))
        print("patching took {} seconds".format(patch_time_elapsed))
        print("stitching took {} seconds".format(stitch_time_elapsed))
        df.loc[idx, "status"] = "processed"

        seg_times += seg_time_elapsed
        patch_times += patch_time_elapsed
        stitch_times += stitch_time_elapsed

    seg_times /= total
    patch_times /= total
    stitch_times /= total

    df.to_csv(os.path.join(save_dir, "process_list_autogen.csv"), index=False)
    print("average segmentation time in s per slide: {}".format(seg_times))
    print("average patching time in s per slide: {}".format(patch_times))
    print("average stiching time in s per slide: {}".format(stitch_times))

    return seg_times, patch_times


def create_patches(
    source: str,
    patch_size: int,
    patch_spacing: float,
    save_dir: str,
    step_size: Optional[int] = None,
    patch: bool = True,
    seg: bool = True,
    stitch: bool = True,
    no_auto_skip: bool = True,
    preset=None,
    process_list=None,
):
    patch_save_dir = os.path.join(save_dir, "patches")
    mask_save_dir = os.path.join(save_dir, "masks")
    stitch_save_dir = os.path.join(save_dir, "stitches")

    if process_list:
        process_list = os.path.join(save_dir, process_list)

    else:
        process_list = None

    print("source: ", source)
    print("patch_save_dir: ", patch_save_dir)
    print("mask_save_dir: ", mask_save_dir)
    print("stitch_save_dir: ", stitch_save_dir)

    directories = {
        "source": source,
        "save_dir": save_dir,
        "patch_save_dir": patch_save_dir,
        "mask_save_dir": mask_save_dir,
        "stitch_save_dir": stitch_save_dir,
    }

    for key, val in directories.items():
        print("{} : {}".format(key, val))
        if key not in ["source"]:
            os.makedirs(val, exist_ok=True)

    seg_params = {
        "seg_level": -1,
        "sthresh": 8,
        "mthresh": 7,
        "close": 4,
        "use_otsu": False,
        "keep_ids": "none",
        "exclude_ids": "none",
    }
    filter_params = {"a_t": 100, "a_h": 16, "max_n_holes": 8}
    vis_params = {"vis_level": -1, "line_thickness": 250}
    patch_params = {"use_padding": True, "contour_fn": "four_pt"}

    if preset:
        preset_df = pd.read_csv(_script_path / "presets" / preset)
        for key in seg_params.keys():
            seg_params[key] = preset_df.loc[0, key]

        for key in filter_params.keys():
            filter_params[key] = preset_df.loc[0, key]

        for key in vis_params.keys():
            vis_params[key] = preset_df.loc[0, key]

        for key in patch_params.keys():
            patch_params[key] = preset_df.loc[0, key]

    parameters = {
        "seg_params": seg_params,
        "filter_params": filter_params,
        "patch_params": patch_params,
        "vis_params": vis_params,
    }

    print(parameters)

    seg_and_patch(
        **directories,
        **parameters,
        patch_size=patch_size,
        step_size=step_size,
        seg=seg,
        use_default_params=False,
        save_mask=True,
        stitch=stitch,
        patch_level=0,  # args.patch_level,
        patch=patch,
        process_list=process_list,
        auto_skip=no_auto_skip,
        patch_spacing=patch_spacing,
    )



"""WSInfer is a toolkit for fast patch-based inference on whole slide images."""

from . import _version
from ._modellib.models import get_model_weights  # noqa
from ._modellib.models import list_all_models_and_weights  # noqa
from ._modellib.models import register_model_weights  # noqa
from ._modellib.run_inference import run_inference  # noqa
from ._modellib.run_inference import WholeSlideImagePatches  # noqa
from ._modellib.transforms import PatchClassification  # noqa

__version__ = _version.get_versions()["version"]

del _version



"""Convert CSVs of model outputs to GeoJSON files.

GeoJSON files can be loaded into whole slide image viewers like QuPath.
"""

import json
from pathlib import Path
import typing

import click
import pandas as pd
import tqdm


def _box_to_polygon(
    *, minx: int, miny: int, width: int, height: int
) -> typing.List[typing.Tuple[int, int]]:
    """Get coordinates of a box polygon."""
    maxx = minx + width
    maxy = miny + height
    return [(maxx, miny), (maxx, maxy), (minx, maxy), (minx, miny), (maxx, miny)]


def _row_to_geojson(row: pd.Series, prob_cols: typing.List[str]) -> typing.Dict:
    """Convert information about one tile to a single GeoJSON feature."""
    minx, miny, width, height = row["minx"], row["miny"], row["width"], row["height"]
    coords = _box_to_polygon(minx=minx, miny=miny, width=width, height=height)
    prob_dict = row[prob_cols].to_dict()
    measurements = [{"name": k, "value": v} for k, v in prob_dict.items()]
    return {
        "type": "Feature",
        "id": "PathTileObject",
        "geometry": {
            "type": "Polygon",
            "coordinates": [coords],
        },
        "properties": {
            "isLocked": True,
            # measurements is a list of {"name": str, "value": float} dicts.
            # https://qupath.github.io/javadoc/docs/qupath/lib/measurements/MeasurementList.html
            "measurements": measurements,
            # classification is a dict of "name": str and optionally "color": int.
            # https://qupath.github.io/javadoc/docs/qupath/lib/objects/classes/PathClass.html
            # We do not include classification because we do not enforce a single class
            # per tile.
            # "classification": {"name": class_name},
        },
    }


def _dataframe_to_geojson(df: pd.DataFrame, prob_cols: typing.List[str]) -> typing.Dict:
    """Convert a dataframe of tiles to GeoJSON format."""
    features = df.apply(_row_to_geojson, axis=1, prob_cols=prob_cols)
    return {
        "type": "FeatureCollection",
        "features": features.tolist(),
    }


def convert(input, output) -> None:
    df = pd.read_csv(input)
    prob_cols = [col for col in df.columns.tolist() if col.startswith("prob_")]
    if not prob_cols:
        raise click.ClickException("Did not find any columns with prob_ prefix.")
    geojson = _dataframe_to_geojson(df, prob_cols)
    with open(output, "w") as f:
        json.dump(geojson, f)


@click.command()
@click.argument(
    "results_dir",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=Path, resolve_path=True
    ),
)
@click.argument(
    "output",
    type=click.Path(exists=False, path_type=Path, resolve_path=True),
)
def cli(*, results_dir: Path, output: Path):
    """Convert model outputs to GeoJSON format.

    GeoJSON files can be used with pathology viewers like QuPath.

    RESULTS_DIR     Path to results directory (containing model-outputs dir).

    OUTPUT          Path to output directory in which to save GeoJSON files.
    """
    if not results_dir.exists():
        raise click.ClickException(f"results_dir does not exist: {results_dir}")
    if output.exists():
        raise click.ClickException("Output directory already exists.")
    if (
        not (results_dir / "model-outputs").exists()
        and (results_dir / "patches").exists()
    ):
        raise click.ClickException(
            "Model outputs have not been generated yet. Please run model inference."
        )
    if not (results_dir / "model-outputs").exists():
        raise click.ClickException(
            "Expected results_dir to contain a 'model-outputs' directory but it does"
            " not. Please provide the path to the directory that contains"
            " model-outputs, masks, and patches."
        )

    csvs = list((results_dir / "model-outputs").glob("*.csv"))
    if not csvs:
        raise click.ClickException("No CSVs found. Did you generate model outputs?")

    output.mkdir(exist_ok=False)

    for input_csv in tqdm.tqdm(csvs):
        output_path = output / input_csv.with_suffix(".json").name
        convert(input=input_csv, output=output_path)

    click.secho(f"Saved outputs to {output}", fg="green")



import h5py
import numpy as np
import os
from PIL import Image
import math
import cv2

from .util_classes import Mosaic_Canvas


def isWhitePatch(patch, satThresh=5):
    patch_hsv = cv2.cvtColor(patch, cv2.COLOR_RGB2HSV)
    return True if np.mean(patch_hsv[:, :, 1]) < satThresh else False


def isBlackPatch(patch, rgbThresh=40):
    return True if np.all(np.mean(patch, axis=(0, 1)) < rgbThresh) else False


def isBlackPatch_S(patch, rgbThresh=20, percentage=0.05):
    num_pixels = patch.size[0] * patch.size[1]
    return (
        True
        if np.all(np.array(patch) < rgbThresh, axis=(2)).sum() > num_pixels * percentage
        else False
    )


def isWhitePatch_S(patch, rgbThresh=220, percentage=0.2):
    num_pixels = patch.size[0] * patch.size[1]
    return (
        True
        if np.all(np.array(patch) > rgbThresh, axis=(2)).sum() > num_pixels * percentage
        else False
    )


def coord_generator(x_start, x_end, x_step, y_start, y_end, y_step, args_dict=None):
    for x in range(x_start, x_end, x_step):
        for y in range(y_start, y_end, y_step):
            if args_dict is not None:
                process_dict = args_dict.copy()
                process_dict.update({"pt": (x, y)})
                yield process_dict
            else:
                yield (x, y)


def savePatchIter_bag_hdf5(patch):
    (
        x,
        y,
        cont_idx,
        patch_level,
        downsample,
        downsampled_level_dim,
        level_dim,
        img_patch,
        name,
        save_path,
    ) = tuple(patch.values())
    img_patch = np.array(img_patch)[np.newaxis, ...]
    img_shape = img_patch.shape

    file_path = os.path.join(save_path, name) + ".h5"
    file = h5py.File(file_path, "a")

    dset = file["imgs"]
    dset.resize(len(dset) + img_shape[0], axis=0)
    dset[-img_shape[0] :] = img_patch

    if "coords" in file:
        coord_dset = file["coords"]
        coord_dset.resize(len(coord_dset) + img_shape[0], axis=0)
        coord_dset[-img_shape[0] :] = (x, y)

    file.close()


def save_hdf5(output_path, asset_dict, attr_dict=None, mode="a"):
    file = h5py.File(output_path, mode)
    for key, val in asset_dict.items():
        data_shape = val.shape
        if key not in file:
            data_type = val.dtype
            chunk_shape = (1,) + data_shape[1:]
            maxshape = (None,) + data_shape[1:]
            dset = file.create_dataset(
                key,
                shape=data_shape,
                maxshape=maxshape,
                chunks=chunk_shape,
                dtype=data_type,
            )
            dset[:] = val
            if attr_dict is not None:
                if key in attr_dict.keys():
                    for attr_key, attr_val in attr_dict[key].items():
                        dset.attrs[attr_key] = attr_val
        else:
            dset = file[key]
            dset.resize(len(dset) + data_shape[0], axis=0)
            dset[-data_shape[0] :] = val
    file.close()
    return output_path


def initialize_hdf5_bag(first_patch, save_coord=False):
    (
        x,
        y,
        cont_idx,
        patch_level,
        downsample,
        downsampled_level_dim,
        level_dim,
        img_patch,
        name,
        save_path,
    ) = tuple(first_patch.values())
    file_path = os.path.join(save_path, name) + ".h5"
    file = h5py.File(file_path, "w")
    img_patch = np.array(img_patch)[np.newaxis, ...]
    dtype = img_patch.dtype

    # Initialize a resizable dataset to hold the output
    img_shape = img_patch.shape
    maxshape = (None,) + img_shape[
        1:
    ]  # maximum dimensions up to which dataset maybe resized (None means unlimited)
    dset = file.create_dataset(
        "imgs", shape=img_shape, maxshape=maxshape, chunks=img_shape, dtype=dtype
    )

    dset[:] = img_patch
    dset.attrs["patch_level"] = patch_level
    dset.attrs["wsi_name"] = name
    dset.attrs["downsample"] = downsample
    dset.attrs["level_dim"] = level_dim
    dset.attrs["downsampled_level_dim"] = downsampled_level_dim

    if save_coord:
        coord_dset = file.create_dataset(
            "coords", shape=(1, 2), maxshape=(None, 2), chunks=(1, 2), dtype=np.int32
        )
        coord_dset[:] = (x, y)

    file.close()
    return file_path


def sample_indices(
    scores, k, start=0.48, end=0.52, convert_to_percentile=False, seed=1
):
    np.random.seed(seed)
    if convert_to_percentile:
        end_value = np.quantile(scores, end)
        start_value = np.quantile(scores, start)
    else:
        end_value = end
        start_value = start
    score_window = np.logical_and(scores >= start_value, scores <= end_value)
    indices = np.where(score_window)[0]
    if len(indices) < 1:
        return -1
    else:
        return np.random.choice(indices, min(k, len(indices)), replace=False)


def top_k(scores, k, invert=False):
    if invert:
        top_k_ids = scores.argsort()[:k]
    else:
        top_k_ids = scores.argsort()[::-1][:k]
    return top_k_ids


def to_percentiles(scores):
    from scipy.stats import rankdata

    scores = rankdata(scores, "average") / len(scores) * 100
    return scores


def screen_coords(scores, coords, top_left, bot_right):
    bot_right = np.array(bot_right)
    top_left = np.array(top_left)
    mask = np.logical_and(
        np.all(coords >= top_left, axis=1), np.all(coords <= bot_right, axis=1)
    )
    scores = scores[mask]
    coords = coords[mask]
    return scores, coords


def sample_rois(
    scores,
    coords,
    k=5,
    mode="range_sample",
    seed=1,
    score_start=0.45,
    score_end=0.55,
    top_left=None,
    bot_right=None,
):

    if len(scores.shape) == 2:
        scores = scores.flatten()

    scores = to_percentiles(scores)
    if top_left is not None and bot_right is not None:
        scores, coords = screen_coords(scores, coords, top_left, bot_right)

    if mode == "range_sample":
        sampled_ids = sample_indices(
            scores,
            start=score_start,
            end=score_end,
            k=k,
            convert_to_percentile=False,
            seed=seed,
        )
    elif mode == "topk":
        sampled_ids = top_k(scores, k, invert=False)
    elif mode == "reverse_topk":
        sampled_ids = top_k(scores, k, invert=True)
    else:
        raise NotImplementedError
    coords = coords[sampled_ids]
    scores = scores[sampled_ids]

    asset = {"sampled_coords": coords, "sampled_scores": scores}
    return asset


def DrawGrid(img, coord, shape, thickness=2, color=(0, 0, 0, 255)):
    cv2.rectangle(
        img,
        tuple(np.maximum([0, 0], coord - thickness // 2)),
        tuple(coord - thickness // 2 + np.array(shape)),
        (0, 0, 0, 255),
        thickness=thickness,
    )
    return img


def DrawMap(
    canvas, patch_dset, coords, patch_size, indices=None, verbose=1, draw_grid=True
):
    if indices is None:
        indices = np.arange(len(coords))
    total = len(indices)
    if verbose > 0:
        ten_percent_chunk = math.ceil(total * 0.1)
        print("start stitching {}".format(patch_dset.attrs["wsi_name"]))

    for idx in range(total):
        if verbose > 0:
            if idx % ten_percent_chunk == 0:
                print("progress: {}/{} stitched".format(idx, total))

        patch_id = indices[idx]
        patch = patch_dset[patch_id]
        patch = cv2.resize(patch, patch_size)
        coord = coords[patch_id]
        canvas_crop_shape = canvas[
            coord[1] : coord[1] + patch_size[1], coord[0] : coord[0] + patch_size[0], :3
        ].shape[:2]
        canvas[
            coord[1] : coord[1] + patch_size[1], coord[0] : coord[0] + patch_size[0], :3
        ] = patch[: canvas_crop_shape[0], : canvas_crop_shape[1], :]
        if draw_grid:
            DrawGrid(canvas, coord, patch_size)

    return Image.fromarray(canvas)


def DrawMapFromCoords(
    canvas,
    wsi_object,
    coords,
    patch_size,
    vis_level,
    indices=None,
    verbose=1,
    draw_grid=True,
):
    downsamples = wsi_object.wsi.level_downsamples[vis_level]
    if indices is None:
        indices = np.arange(len(coords))
    total = len(indices)
    if verbose > 0:
        ten_percent_chunk = math.ceil(total * 0.1)

    patch_size = tuple(
        np.ceil((np.array(patch_size) / np.array(downsamples))).astype(np.int32)
    )
    print("downscaled patch size: {}x{}".format(patch_size[0], patch_size[1]))

    for idx in range(total):
        if verbose > 0:
            if idx % ten_percent_chunk == 0:
                print("progress: {}/{} stitched".format(idx, total))

        patch_id = indices[idx]
        coord = coords[patch_id]
        patch = np.array(
            wsi_object.wsi.read_region(tuple(coord), vis_level, patch_size).convert(
                "RGB"
            )
        )
        coord = np.ceil(coord / downsamples).astype(np.int32)
        canvas_crop_shape = canvas[
            coord[1] : coord[1] + patch_size[1], coord[0] : coord[0] + patch_size[0], :3
        ].shape[:2]
        canvas[
            coord[1] : coord[1] + patch_size[1], coord[0] : coord[0] + patch_size[0], :3
        ] = patch[: canvas_crop_shape[0], : canvas_crop_shape[1], :]
        if draw_grid:
            DrawGrid(canvas, coord, patch_size)

    return Image.fromarray(canvas)


def StitchPatches(
    hdf5_file_path, downscale=16, draw_grid=False, bg_color=(0, 0, 0), alpha=-1
):
    file = h5py.File(hdf5_file_path, "r")
    dset = file["imgs"]
    coords = file["coords"][:]
    if "downsampled_level_dim" in dset.attrs.keys():
        w, h = dset.attrs["downsampled_level_dim"]
    else:
        w, h = dset.attrs["level_dim"]
    print("original size: {} x {}".format(w, h))
    w = w // downscale
    h = h // downscale
    coords = (coords / downscale).astype(np.int32)
    print("downscaled size for stiching: {} x {}".format(w, h))
    print("number of patches: {}".format(len(dset)))
    img_shape = dset[0].shape
    print("patch shape: {}".format(img_shape))
    downscaled_shape = (img_shape[1] // downscale, img_shape[0] // downscale)

    if w * h > Image.MAX_IMAGE_PIXELS:
        raise Image.DecompressionBombError(
            "Visualization Downscale %d is too large" % downscale
        )

    if alpha < 0 or alpha == -1:
        heatmap = Image.new(size=(w, h), mode="RGB", color=bg_color)
    else:
        heatmap = Image.new(
            size=(w, h), mode="RGBA", color=bg_color + (int(255 * alpha),)
        )

    heatmap = np.array(heatmap)
    heatmap = DrawMap(
        heatmap, dset, coords, downscaled_shape, indices=None, draw_grid=draw_grid
    )

    file.close()
    return heatmap


def StitchCoords(
    hdf5_file_path,
    wsi_object,
    downscale=16,
    draw_grid=False,
    bg_color=(0, 0, 0),
    alpha=-1,
):
    wsi = wsi_object.getOpenSlide()
    vis_level = wsi.get_best_level_for_downsample(downscale)
    file = h5py.File(hdf5_file_path, "r")
    dset = file["coords"]
    coords = dset[:]
    w, h = wsi.level_dimensions[0]

    print("start stitching {}".format(dset.attrs["name"]))
    print("original size: {} x {}".format(w, h))

    w, h = wsi.level_dimensions[vis_level]

    print("downscaled size for stiching: {} x {}".format(w, h))
    print("number of patches: {}".format(len(coords)))

    patch_size = dset.attrs["patch_size"]
    patch_level = dset.attrs["patch_level"]
    print(
        "patch size: {}x{} patch level: {}".format(patch_size, patch_size, patch_level)
    )
    patch_size = tuple(
        (
            np.array((patch_size, patch_size)) * wsi.level_downsamples[patch_level]
        ).astype(np.int32)
    )
    print("ref patch size: {}x{}".format(patch_size, patch_size))

    if w * h > Image.MAX_IMAGE_PIXELS:
        raise Image.DecompressionBombError(
            "Visualization Downscale %d is too large" % downscale
        )

    if alpha < 0 or alpha == -1:
        heatmap = Image.new(size=(w, h), mode="RGB", color=bg_color)
    else:
        heatmap = Image.new(
            size=(w, h), mode="RGBA", color=bg_color + (int(255 * alpha),)
        )

    heatmap = np.array(heatmap)
    heatmap = DrawMapFromCoords(
        heatmap,
        wsi_object,
        coords,
        patch_size,
        vis_level,
        indices=None,
        draw_grid=draw_grid,
    )

    file.close()
    return heatmap


def SamplePatches(
    coords_file_path,
    save_file_path,
    wsi_object,
    patch_level=0,
    custom_downsample=1,
    patch_size=256,
    sample_num=100,
    seed=1,
    stitch=True,
    verbose=1,
    mode="w",
):
    file = h5py.File(coords_file_path, "r")
    dset = file["coords"]
    coords = dset[:]

    h5_patch_size = dset.attrs["patch_size"]
    h5_patch_level = dset.attrs["patch_level"]

    if verbose > 0:
        print("in .h5 file: total number of patches: {}".format(len(coords)))
        print(
            "in .h5 file: patch size: {}x{} patch level: {}".format(
                h5_patch_size, h5_patch_size, h5_patch_level
            )
        )

    if patch_level < 0:
        patch_level = h5_patch_level

    if patch_size < 0:
        patch_size = h5_patch_size

    np.random.seed(seed)
    indices = np.random.choice(
        np.arange(len(coords)), min(len(coords), sample_num), replace=False
    )

    target_patch_size = np.array([patch_size, patch_size])

    if custom_downsample > 1:
        target_patch_size = (
            np.array([patch_size, patch_size]) / custom_downsample
        ).astype(np.int32)

    if stitch:
        canvas = Mosaic_Canvas(
            patch_size=target_patch_size[0],
            n=sample_num,
            downscale=4,
            n_per_row=10,
            bg_color=(0, 0, 0),
            alpha=-1,
        )
    else:
        canvas = None

    for idx in indices:
        coord = coords[idx]
        patch = wsi_object.wsi.read_region(
            coord, patch_level, tuple([patch_size, patch_size])
        ).convert("RGB")
        if custom_downsample > 1:
            patch = patch.resize(tuple(target_patch_size))

        # if isBlackPatch_S(patch, rgbThresh=20, percentage=0.05) or isWhitePatch_S(
        # patch, rgbThresh=220, percentage=0.25):
        #     continue

        if stitch:
            canvas.paste_patch(patch)

        asset_dict = {"imgs": np.array(patch)[np.newaxis, ...], "coords": coord}
        save_hdf5(save_file_path, asset_dict, mode=mode)
        mode = "a"

    return canvas, len(coords), len(indices)



"""Create a dense grid of patch coordinates. This does *not* create a tissue mask."""

import itertools
from pathlib import Path
from typing import Tuple

import h5py
import large_image
import numpy as np


def _get_dense_grid(
    slide, orig_patch_size: int, patch_spacing_um_px: float
) -> Tuple[np.ndarray, int]:
    ts = large_image.getTileSource(slide)
    patch_spacing_mm_px = patch_spacing_um_px / 1000
    patch_size = orig_patch_size * patch_spacing_mm_px / ts.getMetadata()["mm_x"]
    patch_size = round(patch_size)
    step_size = patch_size  # non-overlapping patches
    cols = ts.getMetadata()["sizeX"]
    rows = ts.getMetadata()["sizeY"]
    xs = range(0, cols, step_size)
    ys = range(0, rows, step_size)
    # List of (x, y) coordinates.
    return np.asarray(list(itertools.product(xs, ys))), patch_size


def create_grid_and_save(
    slide, results_dir, orig_patch_size: int, patch_spacing_um_px: float
):
    """Create dense grid of (x,y) coordinates and save to HDF5.

    This is similar to the CLAM coordinate code but does not use a tissue mask.
    """
    slide = Path(slide)
    results_dir = Path(results_dir)
    hdf5_path = results_dir / "patches" / f"{slide.stem}.h5"
    hdf5_path.parent.mkdir(exist_ok=True)
    coords, patch_size = _get_dense_grid(
        slide=slide,
        orig_patch_size=orig_patch_size,
        patch_spacing_um_px=patch_spacing_um_px,
    )
    with h5py.File(hdf5_path, "w") as f:
        dset = f.create_dataset("/coords", data=coords, compression="gzip")
        dset.attrs["name"] = str(hdf5_path.stem)
        dset.attrs["patch_level"] = 0
        dset.attrs["patch_size"] = patch_size
        dset.attrs["save_path"] = str(hdf5_path.parent)


def create_grid_and_save_multi_slides(
    wsi_dir, results_dir, orig_patch_size: int, patch_spacing_um_px: float
):
    wsi_dir = Path(wsi_dir)
    slides = list(wsi_dir.glob("*"))
    if not slides:
        raise FileNotFoundError("no slides found")

    for slide in slides:
        create_grid_and_save(
            slide=slide,
            results_dir=results_dir,
            orig_patch_size=orig_patch_size,
            patch_spacing_um_px=patch_spacing_um_px,
        )



"""Run inference.

From the original paper (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7369575/):
> In the prediction (test) phase, no data augmentation was applied except for the
> normalization of the color channels.
"""

from pathlib import Path
import typing
import warnings

import h5py
import large_image
import numpy as np
import pandas as pd
from PIL import Image

# Torch and torchvision are not in the requirements (setup.cfg) because the installation
# can vary based on platform and hardware.
try:
    import torch
except ImportError:
    raise ImportError(
        "Please install torch and torchvision: https://pytorch.org/get-started/locally/"
    )
import tqdm

from .models import Weights

PathType = typing.Union[str, Path]


class WholeSlideImageDirectoryNotFound(FileNotFoundError):
    ...


class WholeSlideImagesNotFound(FileNotFoundError):
    ...


class ResultsDirectoryNotFound(FileNotFoundError):
    ...


class PatchDirectoryNotFound(FileNotFoundError):
    ...


class PatchFilesNotFoundWarning(UserWarning):
    ...


def _read_patch_coords(path: PathType) -> np.ndarray:
    """Read HDF5 file of patch coordinates are return numpy array.

    Returned array has shape (num_patches, 4). Each row has values
    [minx, miny, width, height].
    """
    with h5py.File(path, mode="r") as f:
        coords = f["/coords"][()]
        coords_metadata = f["/coords"].attrs
        if "patch_level" not in coords_metadata.keys():
            raise KeyError(
                "Could not find required key 'patch_level' in hdf5 of patch "
                "coordinates. Has the version of CLAM been updated?"
            )
        patch_level = coords_metadata["patch_level"]
        if patch_level != 0:
            raise NotImplementedError(
                f"This script is designed for patch_level=0 but got {patch_level}"
            )
        if coords.ndim != 2:
            raise ValueError(f"expected coords to have 2 dimensions, got {coords.ndim}")
        if coords.shape[1] != 2:
            raise ValueError(
                f"expected second dim of coords to have len 2 but got {coords.shape[1]}"
            )

        if "patch_size" not in coords_metadata.keys():
            raise KeyError("expected key 'patch_size' in attrs of coords dataset")
        # Append width and height values to the coords, so now each row is
        # [minx, miny, width, height]
        wh = np.full_like(coords, coords_metadata["patch_size"])
        coords = np.concatenate((coords, wh), axis=1)

    return coords


class WholeSlideImagePatches(torch.utils.data.Dataset):
    """Dataset of one whole slide image.

    This object retrieves patches from a whole slide image on the fly.

    Parameters
    ----------
    wsi_path : str, Path
        Path to whole slide image file.
    patch_path : str, Path
        Path to npy file with coordinates of input image.
    um_px : float
        Scale of the resulting patches. Use 0.5 for 20x magnification.
    transform : callable, optional
        A callable to modify a retrieved patch. The callable must accept a
        PIL.Image.Image instance and return a torch.Tensor.
    """

    def __init__(
        self,
        wsi_path: PathType,
        patch_path: PathType,
        um_px: float,
        transform: typing.Optional[typing.Callable[[Image.Image], torch.Tensor]] = None,
    ):
        self.wsi_path = wsi_path
        self.patch_path = patch_path
        self.um_px = float(um_px)
        self.transform = transform

        assert Path(wsi_path).exists(), "wsi path not found"
        assert Path(patch_path).exists(), "patch path not found"

        self.tilesource: large_image.tilesource.TileSource = large_image.getTileSource(
            self.wsi_path
        )
        # Disable the tile cache. We wrap this in a try-except because we are accessing
        # a private attribute. It is possible that this attribute will change names
        # in the future, and if that happens, we do not want to raise errors.
        try:
            self.tilesource.cache._Cache__maxsize = 0
        except AttributeError:
            pass

        self.patches = _read_patch_coords(self.patch_path)
        assert self.patches.ndim == 2, "expected 2D array of patch coordinates"
        # x, y, width, height
        assert self.patches.shape[1] == 4, "expected second dimension to have len 4"

    def __len__(self):
        return self.patches.shape[0]

    def __getitem__(
        self, idx: int
    ) -> typing.Tuple[typing.Union[Image.Image, torch.Tensor], torch.Tensor]:
        coords: typing.Sequence[int] = self.patches[idx]
        assert len(coords) == 4, "expected 4 coords (minx, miny, width, height)"
        minx, miny, width, height = coords
        source_region = dict(
            left=minx, top=miny, width=width, height=height, units="base_pixels"
        )
        target_scale = dict(mm_x=self.um_px / 1000)

        patch_im, _ = self.tilesource.getRegionAtAnotherScale(
            sourceRegion=source_region,
            targetScale=target_scale,
            format=large_image.tilesource.TILE_FORMAT_PIL,
        )
        patch_im = patch_im.convert("RGB")
        if self.transform is not None:
            patch_im = self.transform(patch_im)
        if not isinstance(patch_im, (Image.Image, torch.Tensor)):
            raise TypeError(
                f"patch image must be an Image of Tensor, but got {type(patch_im)}"
            )
        return patch_im, torch.as_tensor([minx, miny, width, height])


def run_inference(
    wsi_dir: PathType,
    results_dir: PathType,
    weights: Weights,
    batch_size: int = 32,
    num_workers: int = 0,
) -> None:
    """Run model inference on a directory of whole slide images and save results to CSV.

    This assumes the patching has already been done and the results are stored in
    `results_dir`. An error will be raised otherwise.

    Output CSV files are written to `{results_dir}/model-outputs/`.

    Parameters
    ----------
    wsi_dir : str or Path
        Directory containing whole slide images. This directory can *only* contain
        whole slide images. Otherwise, an error will be raised during model inference.
    results_dir : str or Path
        Directory containing results of patching.
    weights : wsinfer._modellib.models.Weights
        Instance of Weights including the model object and information about how to
        apply the model to new data.
    batch_size : int
        The batch size during the forward pass (default is 32).
    num_workers : int
        Number of workers for data loading (default is 0, meaning use a single thread).

    Returns
    -------
    None
    """
    # Make sure required directories exist.
    wsi_dir = Path(wsi_dir)
    if not wsi_dir.exists():
        raise WholeSlideImageDirectoryNotFound(f"directory not found: {wsi_dir}")
    wsi_paths = list(wsi_dir.glob("*"))
    if not wsi_paths:
        raise WholeSlideImagesNotFound(wsi_dir)
    results_dir = Path(results_dir)
    if not results_dir.exists():
        raise ResultsDirectoryNotFound(results_dir)

    # Check patches directory.
    patch_dir = results_dir / "patches"
    if not patch_dir.exists():
        raise PatchDirectoryNotFound("Results dir must include 'patches' dir")
    # Create the patch paths based on the whole slide image paths. In effect, only
    # create patch paths if the whole slide image patch exists.
    patch_paths = [patch_dir / p.with_suffix(".h5").name for p in wsi_paths]
    patch_paths_notfound = [p for p in patch_paths if not p.exists()]
    if patch_paths_notfound:
        warnings.warn(
            "Patch extraction seems to have failed for the following slides:"
            + " ".join(str(p.stem) for p in patch_paths_notfound),
            category=PatchFilesNotFoundWarning,
        )

    model_output_dir = results_dir / "model-outputs"
    model_output_dir.mkdir(exist_ok=True)

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    model = weights.load_model()
    model.eval()
    model.to(device)

    # results_for_all_slides: typing.List[pd.DataFrame] = []
    for i, (wsi_path, patch_path) in enumerate(zip(wsi_paths, patch_paths)):
        print(f"Slide {i+1} of {len(wsi_paths)}")
        print(f" Slide path: {wsi_path}")
        print(f" Patch path: {patch_path}")

        slide_csv_name = Path(wsi_path).with_suffix(".csv").name
        slide_csv = model_output_dir / slide_csv_name
        if slide_csv.exists():
            print("Output CSV exists... skipping.")
            print(slide_csv)
            continue

        if not patch_path.exists():
            print(f"Skipping because patch file not found: {patch_path}")
            continue

        dset = WholeSlideImagePatches(
            wsi_path=wsi_path,
            patch_path=patch_path,
            um_px=weights.spacing_um_px,
            transform=weights.transform,
        )

        loader = torch.utils.data.DataLoader(
            dset,
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
        )

        # Store the coordinates and model probabiltiies of each patch in this slide.
        # This lets us know where the probabiltiies map to in the slide.
        slide_coords: typing.List[np.ndarray] = []
        slide_probs: typing.List[np.ndarray] = []
        for batch_imgs, batch_coords in tqdm.tqdm(loader):
            assert batch_imgs.shape[0] == batch_coords.shape[0], "length mismatch"
            with torch.no_grad():
                logits: torch.Tensor = model(batch_imgs.to(device)).detach().cpu()
            # probs has shape (batch_size, num_classes) or (batch_size,)
            if len(logits.shape) > 1 and logits.shape[1] > 1:
                probs = torch.nn.functional.softmax(logits, dim=1)
            else:
                probs = torch.sigmoid(logits.squeeze(1))
            slide_coords.append(batch_coords.numpy())
            slide_probs.append(probs.numpy())

        slide_coords_arr = np.concatenate(slide_coords, axis=0)
        slide_df = pd.DataFrame(
            dict(
                slide=wsi_path,
                minx=slide_coords_arr[:, 0],
                miny=slide_coords_arr[:, 1],
                width=slide_coords_arr[:, 2],
                height=slide_coords_arr[:, 3],
            )
        )
        slide_probs_arr = np.concatenate(slide_probs, axis=0)
        # Use 'prob-' prefix for all classes. This should make it clearer that the
        # column has probabilities for the class. It also makes it easier for us to
        # identify columns associated with probabilities.
        prob_colnames = [f"prob_{c}" for c in weights.class_names]
        slide_df.loc[:, prob_colnames] = slide_probs_arr
        slide_df.to_csv(slide_csv, index=False)
        print("-" * 40)

    return



import click

from .._modellib import models


@click.command()
def cli():
    """Show all available models and weights."""
    models_weights = models.list_all_models_and_weights()

    weights = [models.get_model_weights(*mw) for mw in models_weights]

    print("+-----------------------------------------------------------+")
    click.secho(
        "| MODEL             WEIGHTS        RESOLUTION               |", bold=True
    )
    print("| ========================================================= |")
    _prev_model = models_weights[0][0]
    for (model_name, weights_name), weight_obj in zip(models_weights, weights):
        if _prev_model != model_name:
            print("| --------------------------------------------------------- |")
        _prev_model = model_name
        r = f"{weight_obj.patch_size_pixels} px @ {weight_obj.spacing_um_px} um/px"
        print(f"| {model_name:<18}{weights_name:<15}{r:<25}|")
    print("+-----------------------------------------------------------+")
